import os

import pytest
import toolconfig

from ctc import spec
from ctc import directory


@pytest.mark.parametrize(
    'test',
    [
        [1, 'mainnet'],
        ['mainnet', 'mainnet'],
        [3, 'ropsten'],
    ],
)
def test_get_network_name(test):
    network_reference, network_name = test

    actual_network_name = directory.get_network_name(network=network_reference)
    assert actual_network_name == network_name


@pytest.mark.parametrize(
    'test',
    [
        [1, 1],
        ['mainnet', 1],
        ['ropsten', 3],
    ],
)
def test_get_chain_id(test):
    network_reference, network_chain_id = test

    actual_network_chain_id = directory.get_network_chain_id(
        network=network_reference
    )
    assert actual_network_chain_id == network_chain_id


def test_get_network_metadata():
    metadata = directory.get_network_metadata(network=1)
    assert metadata['name'] == 'mainnet'
    assert metadata['chain_id'] == 1
    assert metadata['block_explorer'] == 'etherscan.io'


def test_get_networks():
    networks = directory.get_networks()
    for network_metadata in networks.values():
        toolconfig.conforms_to_spec(
            data=network_metadata, spec=spec.NetworkMetadata
        )


def test_get_networks_file_path():
    path = directory.get_networks_file_path(use_default=True)
    assert os.path.isfile(path)




from .. import rpc_constructors
from .. import rpc_request
from .. import rpc_digestors


async def async_eth_get_compilers(provider=None):
    request = rpc_constructors.construct_eth_get_compilers()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_compilers(response=response)


async def async_eth_compile_lll(code, provider=None):
    request = rpc_constructors.construct_eth_compile_lll(code=code)
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_compile_lll(response=response)


async def async_eth_compile_solidity(code, provider=None):
    request = rpc_constructors.construct_eth_compile_solidity(code=code)
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_compile_solidity(response=response)


async def async_eth_compile_serpent(code, provider=None):
    request = rpc_constructors.construct_eth_compile_serpent(code=code)
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_compile_serpent(response=response)




import typing

import ctc.config
from ctc import directory
from ctc import spec
from ctc.toolbox import store_utils


_RawAddressData = typing.Union[spec.Address, spec.AddressMetadata]


def import_addresses(
    addresses: dict[str, _RawAddressData],
    label: str,
    overwrite: bool = False,
    network: typing.Optional[spec.NetworkReference] = None,
    backend: spec.StorageBackend = 'Filesystem',
    verbose: bool = True,
    root: typing.Optional[spec.FilesystemRoot] = None,
) -> None:

    if backend != 'Filesystem':
        raise NotImplementedError('storage backend: ' + str(backend))

    if network is None:
        network = ctc.config.get_default_network()

    # process data
    data = []
    for name, address in addresses.items():
        if isinstance(address, str):
            datum: spec.AddressMetadata = {'name': name, 'address': address.lower()}
        elif isinstance(address, dict):
            datum = {
                'name': address['name'],
                'address': address['address'],
            }

            for key, value in address.items():
                if key in ['name', 'address']:
                    pass
                elif key == 'first_block':
                    datum['first_block'] = typing.cast(typing.Optional[int], value)
                elif key == 'contract_name':
                    datum['contract_name'] = typing.cast(typing.Optional[str], value)
                elif key == 'label':
                    assert value == label
                else:
                    raise Exception('unknown key: ' + str(key))
        else:
            raise Exception('unknown address format: ' + str(type(address)))
        data.append(datum)

    # write to disk
    path = directory.get_filesystem_address_path(
        label=label, network=network, root=root
    )
    store_utils.write_file_data(data=data, path=path, overwrite=overwrite)

    # print summary
    if verbose:
        print('wrote', len(addresses), 'addresses to path: ' + str(path))




import typing

from ctc import spec

from . import chainlink_data


async def async_get_eth_price(
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
):
    return await chainlink_data.async_get_feed_datum(
        feed='ETH_USD',
        normalize=normalize,
        provider=provider,
        block=block,
    )


async def async_get_eth_price_by_block(
    blocks: typing.Sequence[spec.BlockNumberReference],
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
):
    return await chainlink_data.async_get_feed_data(
        feed='ETH_USD',
        normalize=normalize,
        provider=provider,
        blocks=blocks,
    )




import toolstr

from ctc import directory
from ctc import evm
from ctc import rpc
from ctc.cli import cli_utils
from ctc.protocols import chainlink_utils


def get_command_spec():
    return {
        'f': async_chainlink_command,
        'args': [
            {'name': 'feed', 'kwargs': {'nargs': '+'}},
            {'name': '--blocks', 'kwargs': {'nargs': '+'}},
            {'name': '--output', 'kwargs': {'default': 'stdout'}},
            {'name': '--overwrite', 'kwargs': {}},
            {'name': '--provider', 'kwargs': {}},
            {'name': '--all-fields', 'kwargs': {'action': 'store_true'}},
            # {'name': '--no-interpolate', 'kwargs': {}},
        ],
    }


async def async_chainlink_command(feed, blocks, output, overwrite, provider, all_fields):

    feed = '_'.join(feed)

    if all_fields:
        fields = 'full'
    else:
        fields = 'answer'

    if blocks is None:
        await chainlink_utils.async_summarize_feed(feed=feed)
    else:
        if evm.is_address_str(feed):
            feed_address = feed
        elif isinstance(feed, str):
            feed_address = directory.get_oracle_address(
                name=feed, protocol='chainlink'
            )
        else:
            raise Exception('unknown feed specification: ' + str(feed))
        name = await rpc.async_eth_call(feed_address, function_name='description')
        toolstr.print_text_box('Chainlink Feed: ' + name)
        print('- feed address')
        print('- feed:', feed)
        print('- fields:', fields)
        print('- output:', output)

        # if cli_utils.is_block_range(blocks):
        #     start_block, end_block = await cli_utils.async_resolve_block_range(blocks)
        #     block_kwargs = {'start_block': start_block, 'end_block': end_block}
        #     print('- block_range: [' + str(start_block) + ', ' + str(end_block) + ']')
        # else:
        resolved_blocks = await cli_utils.async_resolve_block_sample(blocks)
        block_kwargs = {'blocks': resolved_blocks}
        print('- n_blocks:', len(resolved_blocks))
        print('- min_block:', min(resolved_blocks))
        print('- max_block:', max(resolved_blocks))

        print()

        feed_data = await chainlink_utils.async_get_feed_data(
            feed,
            provider=provider,
            fields=fields,
            **block_kwargs,
        )
        df = feed_data

        cli_utils.output_data(df, output=output, overwrite=overwrite)

    await rpc.async_close_http_session()




from __future__ import annotations

import asyncio
import typing

from ctc import evm
from ctc.protocols import chainlink_utils
from ctc.protocols import fei_utils
from ctc.toolbox import pd_utils

from .. import analytics_spec


async def async_compute_prices(
    blocks: list[int], verbose: bool = False
) -> analytics_spec.MetricGroup:
    feed_data = await chainlink_utils.async_get_feed_data(
        feed='FEI_USD', start_block=blocks[0] - 10000
    )
    feed_data = feed_data / 1e8
    # feed_data = evm.interpolate_block_series(
    #     series=feed_data, end_block=blocks[-1]
    # )
    feed_data = pd_utils.interpolate_series(
        series=feed_data,
        end_index=blocks[-1],
    )
    result = [feed_data[block] for block in blocks]

    return {
        'name': 'Prices',
        'metrics': {
            'usd_fei_chainlink': {
                'name': 'USD_FEI Chainlink',
                'values': result,
                'units': 'USD_FEI',
            },
        },
    }


async def async_compute_pfei_by_platform(
    blocks: list[int], verbose: bool = False
) -> analytics_spec.MetricGroup:

    deposit_balances = await fei_utils.async_get_fei_deposit_balances_by_block(
        blocks=blocks,
    )

    platform_balances = fei_utils.fei_deposits_to_deployments_by_block(
        deposit_balances
    )

    metrics: dict[str, analytics_spec.MetricData] = {}
    for name, values in platform_balances.items():
        metrics[name] = {
            'name': name,
            'values': values,
            'units': 'FEI',
        }

    return {
        'name': 'Protocol FEI By Platform',
        'metrics': metrics,
    }


#
# # dex data
#


async def async_compute_dex_tvls(
    blocks: list[int],
    verbose: bool = False,
) -> analytics_spec.MetricGroup:

    pools = analytics_spec.dex_pools

    names = list(pools.keys())
    addresses = [pool['address'] for pool in pools.values()]

    _chunk_by_blocks = False
    if _chunk_by_blocks:
        tasks: list[asyncio.Task[typing.Union[list[int], list[float]]]] = []
        for block in blocks:
            task = evm.async_get_erc20_balance_of_addresses(
                addresses=addresses,
                token='FEI',
                block=block,
            )
            tasks.append(asyncio.create_task(task))
        results = await asyncio.gather(*tasks)
        results = list(zip(*results))

    else:
        tasks = []
        for address in addresses:
            task = evm.async_get_erc20_balance_of_by_block(
                address=address,
                token='FEI',
                blocks=blocks,
            )
            tasks.append(asyncio.create_task(task))
        results = await asyncio.gather(*tasks)

    tvl_by_pool = dict(zip(names, results))

    tvl_by_pool = {
        pool: {'values': tvl, 'units': 'FEI', 'name': pool}
        for pool, tvl in tvl_by_pool.items()
    }

    # sum by platform
    tvl_by_platform = {}
    for pool_name, pool_info in analytics_spec.dex_pools.items():
        pool_tvl = tvl_by_pool[pool_name]
        platform = pool_info['platform']
        if platform not in tvl_by_platform:
            tvl_by_platform[platform] = pool_tvl
        else:
            assert pool_tvl['units'] == tvl_by_platform[platform]['units']
            tvl_by_platform[platform]['values'] = [
                sum(datapoints)
                for datapoints in zip(
                    pool_tvl['values'],
                    tvl_by_platform[platform]['values'],
                )
            ]

    return {
        'name': 'DEX TVL by Platform',
        'metrics': tvl_by_platform,
    }




def build_network_list() -> None:
    """use https://github.com/ethereum-lists/chains repo to construct list"""
    raise NotImplementedError()


def add_network_to_disk():
    raise NotImplementedError()




import typing

from ctc import rpc
from ctc import spec
from ctc.toolbox import backend_utils

from ... import address_utils
from . import contract_abi_backends


def get_backend_functions():
    return {
        'get': {
            'filesystem': contract_abi_backends.async_get_contract_abi_from_filesystem,
            'download': async_download_contract_abi,
            'etherscan': contract_abi_backends.async_get_contract_abi_from_etherscan,
        },
        'save': {
            'filesystem': contract_abi_backends.async_save_contract_abi_to_filesystem,
        },
    }


async def async_get_contract_abi(**query):
    backend_order = query.get('backend_order')
    if backend_order is None:
        backend_order = ['filesystem', 'download']
        query['backend_order'] = backend_order
    return await backend_utils.async_run_on_backend(
        backend_functions=get_backend_functions()['get'], **query
    )


async def async_save_contract_abi(contract_abi, **query):
    query['contract_abi'] = contract_abi
    return await backend_utils.async_run_on_backend(
        backend_functions=get_backend_functions()['save'], **query
    )


async def async_transfer_contract_abi(**kwargs):
    return await backend_utils.async_transfer_backends(
        get=async_get_contract_abi, save=async_save_contract_abi, **kwargs
    )


async def async_download_contract_abi(
    contract_address,
    name=None,
    provider: spec.ProviderSpec = None,
    **kwargs
):

    provider = rpc.get_provider(provider)
    network = provider['network']
    if network is None:
        raise Exception('could not determine network')

    common_kwargs = {'contract_address': contract_address, 'network': network}
    kwargs.setdefault('common_kwargs', {})
    kwargs['common_kwargs'] = dict(kwargs['common_kwargs'])
    kwargs['common_kwargs'].update(common_kwargs)

    save_kwargs = {'name': name, 'network': network}
    kwargs.setdefault('save_kwargs', {})
    kwargs['save_kwargs'] = dict(kwargs['save_kwargs'])
    kwargs['save_kwargs'].update(save_kwargs)

    contract_abi = await async_transfer_contract_abi(
        from_backend='etherscan', to_backend='filesystem', **kwargs
    )

    try:
        await address_utils.async_save_eip897_abi(
            contract_address=contract_address, provider=provider,
        )
        contract_abi = (
            await contract_abi_backends.async_get_contract_abi_from_filesystem(
                contract_address=contract_address, network=network,
            )
        )
    except Exception:
        pass

    return contract_abi





import toolstr

from ctc import evm
from ctc import rpc


def get_command_spec():
    return {
        'f': async_balance_command,
        'args': [
            {'name': 'address'},
            {'name': '--block'},
            {'name': '--raw', 'kwargs': {'action': 'store_true'}},
        ],
    }


async def async_balance_command(address, block, raw):

    if block is not None:
        block = await evm.async_block_number_to_int(block)

    balance = await evm.async_get_eth_balance(
        address=address,
        block=block,
        normalize=(not raw),
    )
    if raw:
        print(balance)
    else:
        print(toolstr.format(balance))
    await rpc.async_close_http_session()




from ctc import spec
from ctc import binary
from .. import rpc_format


def digest_eth_gas_price(
    response: spec.RpcSingularResponse, decode_response: bool = True
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_accounts(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_sign(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_sign_transaction(
    response: spec.RpcSingularResponse, snake_case_response: bool = True
) -> spec.RpcSingularResponse:
    if snake_case_response:
        response = rpc_format.keys_to_snake_case(response)
    return response


def digest_eth_send_transaction(
    response: spec.RpcSingularResponse, snake_case_response: bool = True
) -> spec.RpcSingularResponse:
    if snake_case_response:
        response = rpc_format.keys_to_snake_case(response)
    return response


def digest_eth_send_raw_transaction(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response




"""
see https://eth.wiki/en/fundamentals/rlp
"""

from . import formats


def rlp_encode(data, output_format='prefix_hex'):
    import rlp  # type: ignore

    rlp_data = rlp.encode(data)
    return formats.convert(rlp_data, output_format)


def rlp_decode(data):
    import rlp  # type: ignore

    binary_data = formats.convert(data, 'binary')

    return rlp.decode(binary_data)


# def rlp_encode(input):
#    if isinstance(input, str):
#        if len(input) == 1 and ord(input) < 0x80:
#            return input
#        else:
#            return encode_length(len(input), 0x80) + input
#    elif isinstance(input, list):
#        output = ''
#        for item in input:
#            output += rlp_encode(item)
#        return encode_length(len(output), 0xC0) + output


# def encode_length(L, offset):
#    if L < 56:
#        return chr(L + offset)
#    elif L < 256 ** 8:
#        BL = to_binary(L)
#        return chr(len(BL) + offset + 55) + BL
#    else:
#        raise Exception("input too long")


# def to_binary(x):
#    if x == 0:
#        return ''
#    else:
#        return to_binary(int(x / 256)) + chr(x % 256)


##
## # decode
##
# def rlp_decode(input):
#    if len(input) == 0:
#        return
#    output = ''
#    (offset, dataLen, type) = decode_length(input)
#    if type is str:
#        output = str(substr(input, offset, dataLen))
#    elif type is list:
#        output = list(substr(input, offset, dataLen))
#    output + rlp_decode(substr(input, offset + dataLen))
#    return output


# def decode_length(input):
#    length = len(input)
#    if length == 0:
#        raise Exception("input is null")
#    prefix = ord(input[0])
#    if prefix <= 0x7F:
#        return (0, 1, str)
#    elif prefix <= 0xB7 and length > prefix - 0x80:
#        strLen = prefix - 0x80
#        return (1, strLen, str)
#    elif (
#        prefix <= 0xBF
#        and length > prefix - 0xB7
#        and length > prefix - 0xB7 + to_integer(substr(input, 1, prefix - 0xB7))
#    ):
#        lenOfStrLen = prefix - 0xB7
#        strLen = to_integer(substr(input, 1, lenOfStrLen))
#        return (1 + lenOfStrLen, strLen, str)
#    elif prefix <= 0xF7 and length > prefix - 0xC0:
#        listLen = prefix - 0xC0
#        return (1, listLen, list)
#    elif (
#        prefix <= 0xFF
#        and length > prefix - 0xF7
#        and length > prefix - 0xF7 + to_integer(substr(input, 1, prefix - 0xF7))
#    ):
#        lenOfListLen = prefix - 0xF7
#        listLen = to_integer(substr(input, 1, lenOfListLen))
#        return (1 + lenOfListLen, listLen, list)
#    else:
#        raise Exception("input don't conform RLP encoding form")


# def to_integer(b):
#    length = len(b)
#    if length == 0:
#        raise Exception("input is null")
#    elif length == 1:
#        return ord(b[0])
#    else:
#        return ord(substr(b, -1)) + to_integer(substr(b, 0, -1)) * 256


# def substr(s, beginning, length=None):
#    if length is None:
#        return s[beginning:]
#    else:
#        return s[beginning: length + 1]




from __future__ import annotations

import os
import typing

import ctc.config
from ctc import spec
from ctc.toolbox import store_utils
from ctc.toolbox import search_utils
from . import networks


def get_address_metadata(
    *,
    label: str,
    address: typing.Optional[spec.Address] = None,
    name: typing.Optional[str] = None,
    contract_name: typing.Optional[str] = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
    network: typing.Optional[spec.NetworkReference] = None,
    backend: spec.StorageBackend = 'Filesystem',
) -> spec.AddressMetadata:

    if network is None:
        network = ctc.config.get_default_network()

    if backend == 'Filesystem':

        # load addresses
        addresses = load_filesystem_label_addresses(
            label=label, network=network
        )

        # build query
        query = {
            'address': address,
            'name': name,
            'contract_name': contract_name,
        }
        query = {k: v for k, v in query.items() if v is not None}

        if block is None:
            return search_utils.get_matching_entry(
                sequence=addresses, query=query
            )
        else:

            # find matches
            matches = search_utils.get_matching_entries(
                sequence=addresses, query=query
            )
            if len(matches) == 0:
                raise LookupError('no address found')
            if any(match['first_block'] is None for match in matches):
                raise LookupError('addresses do not specify block range')

            # cast because not None established above
            first_blocks = typing.cast(
                list[tuple[int, int]],
                [(match['first_block'], m) for m, match in enumerate(matches)],
            )

            # sort matches
            first_blocks = sorted(first_blocks)

            if block == 'latest':
                # case: return latest
                return matches[first_blocks[-1][1]]

            elif isinstance(block, int):
                # case: search for latest valid match
                if block < first_blocks[0][0]:
                    raise LookupError('no matching address for specified block')
                for m in range(1, len(matches)):
                    if block < first_blocks[m][0]:
                        return matches[first_blocks[m - 1][1]]
                else:
                    return matches[first_blocks[-1][1]]

            else:
                raise Exception(
                    'unsupported block specification: ' + str(block)
                )

    else:
        raise NotImplementedError('storage backend: ' + str(backend))


def get_address_first_block(
    *,
    label: str,
    address: typing.Optional[spec.Address] = None,
    name: typing.Optional[str] = None,
    contract_name: typing.Optional[str] = None,
    network: typing.Optional[spec.NetworkReference] = None,
    backend: spec.StorageBackend = 'Filesystem',
) -> int:

    metadata = get_address_metadata(
        address=address,
        name=name,
        contract_name=contract_name,
        label=label,
        network=network,
        backend=backend,
    )

    first_block = metadata.get('first_block')
    if first_block is not None:
        return first_block
    else:
        raise Exception('first block not specified for addresss')


def get_address(
    *,
    label: str,
    address: typing.Optional[spec.Address] = None,
    name: typing.Optional[str] = None,
    contract_name: typing.Optional[str] = None,
    block: spec.BlockNumberReference = None,
    network: typing.Optional[spec.NetworkReference] = None,
    backend: spec.StorageBackend = 'Filesystem',
) -> spec.Address:

    return get_address_metadata(
        address=address,
        name=name,
        contract_name=contract_name,
        block=block,
        label=label,
        network=network,
        backend=backend,
    )['address']


def load_filesystem_label_addresses(
    label: str,
    network: spec.NetworkReference,
) -> list[spec.AddressMetadata]:
    path = get_filesystem_address_path(label=label, network=network)
    return store_utils.load_file_data(path=path)


def get_filesystem_address_path(
    label: str,
    network: spec.NetworkReference,
    root: typing.Optional[spec.FilesystemRoot] = None,
) -> str:

    if root is None:
        root = 'root'

    if root == 'root':
        data_dir = ctc.config.get_data_dir()
    elif root == 'default':
        data_dir = ctc.config.get_default_data_dir()
    else:
        raise Exception()

    network_name = networks.get_network_name(network=network)
    return os.path.join(data_dir, network_name, 'addresses', label + '.csv')




import typing

from ctc import spec

from .. import block_utils
from .. import event_utils
from . import erc20_metadata
from . import erc20_abis


def get_token_amount_column(df):
    if 'arg__value' in df:
        return 'arg__value'
    elif 'arg__amount' in df:
        return 'arg__amount'
    elif 'arg__wad' in df:
        return 'arg__wad'
    else:
        raise Exception('could not detect amount column')


async def async_get_erc20_transfers(
    token_address: spec.ERC20Address,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    **event_kwargs
) -> spec.DataFrame:

    transfers = await event_utils.async_get_events(
        contract_address=token_address,
        event_abi=erc20_abis.erc20_event_abis['Transfer'],
        start_block=start_block,
        end_block=end_block,
        **event_kwargs
    )

    # detect amount column
    column = get_token_amount_column(df=transfers)
    transfers[column] = transfers[column].map(int)

    if normalize:

        # get block
        if start_block is not None:
            block = block_utils.standardize_block_number(start_block)
        elif end_block != 'latest' and end_block is not None:
            block = block_utils.standardize_block_number(end_block)
        else:
            block = await block_utils.async_get_latest_block_number()

        decimals = await erc20_metadata.async_get_erc20_decimals(
            token=token_address, block=block
        )
        dtype = float
        transfers[column] = transfers[column] / dtype('1e' + str(decimals))

    return transfers


async def async_get_erc20_holdings_from_transfers(
    transfers: spec.DataFrame,
    block: typing.Optional[spec.BlockNumberReference] = None,
    dtype: typing.Optional[
        typing.Union[typing.Type[int], typing.Type[float]]
    ] = None,
    normalize: bool = False,
) -> spec.DataFrame:

    # filter block
    if block is not None:
        blocks = transfers.index.get_level_values('block_number').values
        mask = blocks <= block
        transfers = transfers[mask]

    amount_key = get_token_amount_column(transfers)

    # convert to float
    if dtype is not None:
        transfers[amount_key] = transfers[amount_key].map(dtype)

    # subtract transfers out from transfers in
    from_transfers = transfers.groupby('arg__from')[amount_key].sum()
    to_transfers = transfers.groupby('arg__to')[amount_key].sum()
    balances = to_transfers.sub(from_transfers, fill_value=0)

    if normalize:
        decimals = await erc20_metadata.async_get_erc20_decimals(
            transfers['contract_address'].values[0]
        )
        balances /= 10 ** decimals

    # sort
    balances = balances.sort_values(ascending=False)

    balances.name = 'balance'
    balances.index.name = 'address'

    return balances




from ctc import rpc


def get_command_spec():
    return {
        'f': async_call_command,
        'args': [
            {'name': 'to_address'},
            {'name': 'function_name'},
            {'name': 'args', 'kwargs': {'nargs': '*'}},
            {'name': '--quiet', 'kwargs': {'action': 'store_true'}},
        ],
    }


async def async_call_command(to_address, function_name, args, quiet):

    if not quiet:
        print('performing eth_call')
        print('- to address:', to_address)
        print('- function:', function_name)
        print('- arguments:', args)
        print()
        print('result:')

    result = await rpc.async_eth_call(
        to_address=to_address,
        function_name=function_name,
        function_parameters=args,
    )
    print(result)

    await rpc.async_close_http_session()




"""WIP"""

import functools

import pandas as pd

from ctc import binary
from ctc import evm
from ctc.rpc.rpc_executors_async import rpc_trace_executors_async


async def get_transaction_traces_df(transaction_hash, provider=None):
    traces = await rpc_trace_executors_async.async_trace_transaction(
        transaction_hash=transaction_hash,
    )

    data = {}

    for trace in traces:
        for key, value in trace.items():
            if isinstance(value, dict):
                for subkey, subvalue in value.items():
                    full_key = key + '__' + subkey
                    data.setdefault(full_key, [])
                    data[full_key].append(subvalue)
            else:
                data.setdefault(key, [])
                data[key].append(value)

    df = pd.DataFrame(data)

    # convert columns to int
    convert_int_columns = [
        'action__gas',
        'result__gasUsed',
        'action__value',
    ]
    f_to_int = functools.partial(binary.convert, output_format='integer')
    for column in convert_int_columns:
        df[column] = df[column].map(f_to_int)

    # remove redundant columns
    remove_columns = [
        'transactionHash',
        'transactionPosition',
        'blockNumber',
        'blockHash',
    ]
    df = df.drop(remove_columns, axis=1)

    # decode call data and parameters
    await _add_trace_call_data(df)

    return df


async def _add_trace_call_data(df):

    contract_abis = {}
    for contract_address in set(df['action__to']):
        contract_abis[contract_address] = await evm.async_get_contract_abi(
            contract_address=contract_address,
        )

    action_call_data = {
        'action__name': [],
        'action__parameters': [],
        'result__outputs': [],
    }

    for action_to, action_input, result__output in zip(
        df['action__to'], df['action__input'], df['result__output']
    ):

        if action_input == '0x':
            action_call_data['action__name'].append('')
            action_call_data['action__parameters'].append(tuple())
            action_call_data['result__outputs'].append(None)

        else:

            try:

                decoded = binary.decode_call_data(
                    contract_abi=contract_abis[action_to],
                    call_data=action_input,
                )
                action_call_data['action__name'].append(
                    decoded['function_abi']['name']
                )
                action_call_data['action__parameters'].append(
                    decoded['parameters']
                )

                function_abi = binary.get_function_abi(
                    contract_abi=contract_abis[action_to],
                    function_selector=action_input[:10],
                )
                result__outputs = binary.decode_function_output(
                    encoded_output=result__output,
                    function_abi=function_abi,
                )
                action_call_data['result__outputs'].append(result__outputs)

            except LookupError:

                action_call_data['action__name'].append('UNKNOWN')
                action_call_data['action__parameters'].append('UNKNOWN')
                action_call_data['result__outputs'].append('UNKNOWN')

    df['action__name'] = action_call_data['action__name']
    df['action__parameters'] = action_call_data['action__parameters']
    df['result__outputs'] = action_call_data['result__outputs']




import pytest

from ctc.protocols import uniswap_v2_utils


@pytest.mark.asyncio
async def test_fetch_uni_v2_pool_state():
    results = await uniswap_v2_utils.async_get_pool_state(
        pool='0x94B0A3d511b6EcDb17eBF877278Ab030acb0A878',
        # x_address='0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2',
        # y_address='0x956F47F50A910163D8BF957Cf5846D573E7f87CA',
        block=13437866,
        # x_name='WETH',
        # y_name='FEI',
        normalize=False,
    )
    assert results['y_reserves'] == 36632373146254861687157
    assert results['x_reserves'] == 136219847556588434237761989
    assert results['lp_total_supply'] == 2152734217700018553905574




import typing

from typing_extensions import TypedDict


NetworkId = int
NetworkName = str
NetworkReference = typing.Union[NetworkId, NetworkName, None]


class NetworkMetadata(TypedDict):
    name: NetworkName
    chain_id: int
    block_explorer: str  # url




import asyncio
import typing

import pandas as pd

from ctc import spec
from ctc import evm
from ctc.toolbox import pd_utils

from .. import chainlink_spec
from . import feed_datum


async def async_get_feed_answer_datum_by_block(
    feed: chainlink_spec._FeedReference,
    blocks: typing.Sequence[spec.BlockNumberReference],
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
    interpolate: bool = True,
) -> spec.Series:

    int_blocks = await evm.async_block_numbers_to_int(blocks=blocks)

    # query data
    coroutines = []
    for block in int_blocks:
        coroutine = feed_datum.async_get_feed_datum(
            feed=feed,
            fields='answer',
            normalize=normalize,
            block=block,
            provider=provider,
        )
        coroutines.append(coroutine)
    result = await asyncio.gather(*coroutines)

    # create series
    series = pd.Series(data=result, index=int_blocks)

    # interpolate blocks
    if interpolate:
        series = pd_utils.interpolate_series(series=series)

    return series


async def async_get_feed_full_datum_by_block(
    feed: chainlink_spec._FeedReference,
    blocks: typing.Sequence[spec.BlockNumberReference],
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
    interpolate: bool = True,
) -> spec.DataFrame:

    int_blocks = await evm.async_block_numbers_to_int(blocks=blocks)

    # query data
    coroutines = []
    for block in int_blocks:
        coroutine = feed_datum.async_get_feed_datum(
            feed=feed,
            block=block,
            provider=provider,
            normalize=normalize,
            fields='full',
        )
        coroutines.append(coroutine)
    result = await asyncio.gather(*coroutines)

    # create series
    df = pd.DataFrame(result, index=int_blocks)

    # interpolate blocks
    if interpolate:
        df = pd_utils.interpolate_dataframe(df=df)

    return df




from __future__ import annotations

import typing
import os

from ctc import spec
from ctc.toolbox import search_utils
from . import config_read


def get_config_version() -> str:
    return config_read.get_config()['version']


def get_data_dir() -> str:
    data_dir = config_read.get_config()['data_dir']
    data_dir = os.path.expanduser(data_dir)
    return data_dir


def get_reports_dir() -> str:
    return os.path.join(get_data_dir(), 'notebooks')


def get_networks() -> dict[spec.NetworkName, spec.NetworkMetadata]:

    return config_read.get_config()['networks']


def get_providers() -> dict[spec.NetworkName, spec.Provider]:
    return config_read.get_config()['providers']


def has_provider(
    *, name=typing.Optional[str], network=typing.Optional[str]
) -> bool:
    try:
        get_provider(name=name, network=network)
        return True
    except LookupError:
        return False


def get_provider(
    *,
    name: typing.Optional[str] = None,
    network: typing.Optional[str] = None,
    protocol: typing.Optional[str] = None,
) -> spec.Provider:

    providers = list(get_providers().values())

    # build query
    query = {}
    if name is None and network is None:
        raise Exception('specify network name or network')
    if name is not None:
        query['name'] = name
    if network is not None:
        query['network'] = network
    if protocol is not None:
        query['protocol'] = protocol

    return search_utils.get_matching_entry(sequence=providers, query=query)


def get_default_network() -> spec.NetworkName:
    return config_read.get_config()['network_defaults']['default_network']


def get_default_provider(
    network: typing.Optional[spec.NetworkName] = None,
) -> spec.Provider:

    if network is None:
        network = get_default_network()

    config = config_read.get_config()
    default_providers = config['network_defaults']['default_providers']

    if network in default_providers:
        provider_name = default_providers[network]
        return get_provider(name=provider_name)
    else:
        raise Exception(
            'no default provider specified for network ' + str(network)
        )




from __future__ import annotations

import os

import toolcli

from .. import config_data
from .. import config_read


def setup_data_root(styles: dict[str, str]) -> tuple[str, bool]:

    print()
    print()
    toolcli.print('## Data Root Directory', style=styles['header'])

    input_kwargs: toolcli.InputFilenameOrDirectoryKwargs = {
        'prompt': 'Where should ctc store data? (specify a directory path)\n> ',
        'create_directory': 'prompt_and_require',
        'style': styles['question'],
    }

    if config_read.config_path_exists():
        try:
            config = config_read.get_config(validate=False)
            old_data_root = config['data_dir']
        except Exception:
            old_data_root = None
    else:
        old_data_root = None

    # data_root
    if old_data_root is None:
        print()
        new_data_root = toolcli.input_filename_or_directory(**input_kwargs)
        create_new_config = True

    else:
        print()
        toolcli.print(
            'Data directory currently set to:',
            toolcli.add_style(old_data_root, styles['path']),
        )

        if not isinstance(old_data_root, str):
            print()
            print('This value is invalid')
            print()
            new_data_root = toolcli.input_filename_or_directory(**input_kwargs)
            create_new_config = True

        elif os.path.abspath(old_data_root) != old_data_root:
            print()
            print('This path should be an absolute path')
            print()
            print('Absolute path:', os.path.abspath(old_data_root))
            print()
            if toolcli.input_yes_or_no('Use this path for data directory?'):
                new_data_root = old_data_root
                create_new_config = False
            else:
                print()
                new_data_root = toolcli.input_filename_or_directory(
                    **input_kwargs
                )
                create_new_config = True

        else:
            print()
            if toolcli.input_yes_or_no(
                'Keep storing data in this directory? ',
                style=styles['question'],
                default='yes',
            ):
                new_data_root = old_data_root
                create_new_config = False
            else:
                new_data_root = toolcli.input_filename_or_directory(
                    **input_kwargs
                )
                create_new_config = True

    # initialize directory data
    if config_data.is_data_root_initialized(new_data_root):
        print()
        print('Data directory is already initialized')
    else:
        initialized = config_data.initialize_data_root(new_data_root)
        if not initialized:
            return setup_data_root(styles=styles)

    return new_data_root, create_new_config




import asyncio
import json
import os
import time

import toolstr

from ctc import rpc
from ctc.protocols import fei_utils


def get_command_spec():
    return {
        'f': async_payload_command,
        'args': [
            {'name': 'timescale', 'kwargs': {'nargs': '?', 'default': None}},
            {'name': '--path'},
            {'name': '--overwrite', 'kwargs': {'action': 'store_true'}},
        ],
    }


async def async_payload_command(timescale, path, overwrite, **kwargs):

    # validate inputs
    if timescale is None:
        timescale = '30d, 1d'
    timescale = fei_utils.resolve_timescale(timescale)
    if path is None:
        name = 'payload_{window_size}_{interval_size}'.format(**timescale)
        path = './' + name + '.json'

    # print summary
    print('generating data payload')
    print('- interval size:', timescale['interval_size'])
    print('- window size:', timescale['window_size'])
    print('- output path:', path)

    if os.path.exists(path) and not overwrite:
        raise Exception('path already exists: ' + str(path))

    # create payload
    print()
    print('starting...')
    start_time = time.time()
    payload = await fei_utils.async_create_payload(timescale=timescale)
    end_time = time.time()
    print()
    print('...done (t=' + toolstr.format(end_time - start_time) + 's)')

    # save payload
    with open(path, 'w') as f:
        json.dump(payload, f)

    provider = rpc.get_provider()
    await rpc.async_close_http_session(provider=provider)




from __future__ import annotations

import os
import typing

import toolcache

import ctc.config
from ctc import spec
from ctc.toolbox import search_utils
from ctc.toolbox import store_utils
from . import networks


_LabelGroup = typing.Union[str, typing.Sequence[str]]

default_erc20_label = '1inch'


def get_erc20_metadata(
    *,
    symbol: typing.Optional[spec.ERC20Symbol] = None,
    address: typing.Optional[spec.ERC20Address] = None,
    label: typing.Optional[_LabelGroup] = None,
    network: spec.NetworkReference = None,
    backend: spec.StorageBackend = 'Filesystem',
) -> spec.ERC20Metadata:

    if network is None:
        network = ctc.config.get_default_network()

    if backend == 'Filesystem':

        # build query
        query = {}
        if symbol is not None:
            query['symbol'] = symbol
        if address is not None:
            query['address'] = address.lower()

        # get label list
        if label is None:
            label = default_erc20_label
        if isinstance(label, str):
            label_list: _LabelGroup = [label]
        else:
            label_list = label

        # query each label, return the first one that matches
        for query_label in label_list:

            # load label erc20
            erc20_data = load_filesystem_erc20_data(
                network=network,
                label=query_label,
            )

            # return if match found
            try:
                return search_utils.get_matching_entry(
                    erc20_data,
                    query=query,
                )
            except search_utils.NoMatchFound:
                pass

        else:
            raise LookupError('could not find erc20 metadata')

    else:
        raise Exception('unknown backend: ' + str(backend))


def get_erc20_address(
    symbol: typing.Optional[spec.ERC20Symbol],
    *,
    address: typing.Optional[spec.ERC20Address] = None,
    label: typing.Optional[str] = None,
    network: typing.Optional[spec.NetworkReference] = None,
    backend: spec.StorageBackend = 'Filesystem',
) -> spec.ERC20Address:

    if address is not None:
        return address

    metadata = get_erc20_metadata(
        symbol=symbol,
        label=label,
        network=network,
        backend=backend,
    )

    return metadata['address']


def has_erc20_metadata(
    address: typing.Optional[spec.ERC20Address] = None,
    symbol: typing.Optional[spec.ERC20Symbol] = None,
    **kwargs,
) -> bool:

    try:
        get_erc20_metadata(address=address, symbol=symbol, **kwargs)
        return True
    except LookupError:
        return False


def get_erc20_symbol(
    address: typing.Optional[spec.ERC20Address] = None,
    *,
    symbol: typing.Optional[spec.ERC20Symbol] = None,
    label: typing.Optional[str] = None,
    network: typing.Optional[spec.NetworkReference] = None,
    backend: spec.StorageBackend = 'Filesystem',
) -> spec.ERC20Symbol:

    if symbol is not None:
        return symbol

    metadata = get_erc20_metadata(
        label=label,
        address=address,
        backend=backend,
        network=network,
    )

    return metadata['symbol']


#
# # backend-specific
#


@toolcache.cache(cachetype='memory')
def load_filesystem_erc20_data(
    network: spec.NetworkReference,
    label: typing.Optional[str] = None,
) -> list[spec.ERC20Metadata]:

    # set default label
    if label is None:
        label = '1inch'

    # build path
    path = get_erc20_data_path(network=network, label=label)

    # load data
    return store_utils.load_file_data(path)


def get_erc20_data_path(network: spec.NetworkReference, label: str) -> str:
    network_name = networks.get_network_name(network=network)
    data_dir = ctc.config.get_data_dir()
    filename = label + '.csv'
    return os.path.join(data_dir, network_name, 'erc20s', filename)




import numpy as np


def _ensure_exactly_one(*args):
    non_none = list(arg for arg in args if arg is not None)
    if len(non_none) > 1:
        raise Exception('too many arguments specified')
    if len(non_none) == 0:
        raise Exception('not enough arguments specified')
    return non_none[0]


def _ensure_positive(value, error=True):
    if isinstance(value, (np.ndarray, tuple, list)):
        positive = all(subvalue > 0 for subvalue in value)
    else:
        positive = value > 0

    if error and not positive:
        raise Exception('value must be positive')

    return positive


def _ensure_non_negative(value, error=True):
    if isinstance(value, (np.ndarray, tuple, list)):
        non_negative = all(
            subvalue >= 0 or np.isclose(subvalue, 0, atol=1e-7) for subvalue in value
        )
    else:
        non_negative = value >= 0 or np.isclose(value, 0, atol=1e-7)

    if error and not non_negative:
        raise Exception('value must be non negative')

    return non_negative


def _ensure_values_equal(lhs, rhs):
    if isinstance(lhs, dict):
        assert set(lhs.keys()) == set(rhs.keys())
        for key in lhs.keys():
            _ensure_values_equal(lhs[key], rhs[key])
    else:
        assert np.isclose(lhs, rhs)




import asyncio

from ctc import evm
from ctc import rpc
from ctc import spec


#
# # metadata
#


async def download_pool_abi(pool_address):
    example_pool = '0x8f8ef111b67c04eb1641f5ff19ee54cda062f163'
    await evm.async_save_proxy_contract_abi_to_filesystem(
        contract_address=pool_address,
        proxy_implementation=example_pool,
    )


async def ensure_pool_abi(pool_address):
    try:
        await evm.async_get_contract_abi(contract_address=pool_address)
    except spec.AbiNotFoundException:
        await download_pool_abi(pool_address)


async def async_get_pool_tokens(pool_address, **rpc_kwargs):
    kwargs = dict(rpc_kwargs, to_address=pool_address)
    return await asyncio.gather(
        rpc.async_eth_call(function_name='token0', **kwargs),
        rpc.async_eth_call(function_name='token1', **kwargs),
    )


async def async_get_pool_metadata(pool_address, **rpc_kwargs):
    await ensure_pool_abi(pool_address)

    x_address, y_address = await async_get_pool_tokens(
        pool_address=pool_address
    )
    x_symbol, y_symbol = await evm.async_get_erc20s_symbols(
        tokens=[x_address, y_address], **rpc_kwargs
    )
    return {
        'x_symbol': x_symbol,
        'y_symbol': y_symbol,
        'x_address': x_address,
        'y_address': y_address,
    }


#
# # events
#


async def async_get_pool_swaps(
    pool_address,
    start_block=None,
    end_block=None,
    replace_symbols=False,
    normalize=True,
):
    await ensure_pool_abi(pool_address)

    if normalize or replace_symbols:
        metadata_task = asyncio.create_task(
            async_get_pool_metadata(pool_address)
        )

    swaps = await evm.async_get_events(
        event_name='Swap',
        contract_address=pool_address,
        start_block=start_block,
        end_block=end_block,
    )

    if normalize or replace_symbols:
        metadata = await metadata_task

    # rename columns
    if replace_symbols:
        x_symbol = metadata['x_symbol']
        y_symbol = metadata['y_symbol']
    else:
        x_symbol = 'x'
        y_symbol = 'y'
    columns = {
        'arg__amount0': x_symbol + '_amount',
        'arg__amount1': y_symbol + '_amount',
    }
    swaps = swaps.rename(columns=columns)

    # normalize columns
    if normalize:
        x_decimals, y_decimals = await evm.async_get_erc20s_decimals(
            tokens=[metadata['x_address'], metadata['y_address']],
        )
        swaps[columns['arg__amount0']] = swaps[columns['arg__amount0']].astype(
            float
        ) / (10 ** x_decimals)
        swaps[columns['arg__amount1']] = swaps[columns['arg__amount1']].astype(
            float
        ) / (10 ** y_decimals)

    return swaps




from ctc import spec
from .. import rpc_request


def construct_eth_get_compilers() -> spec.RpcRequest:
    return rpc_request.create('eth_getCompilers', [])


def construct_eth_compile_lll(code: str) -> spec.RpcRequest:
    return rpc_request.create('eth_compileLLL', [code])


def construct_eth_compile_solidity(code: str) -> spec.RpcRequest:
    return rpc_request.create('eth_compileSolidity', [code])


def construct_eth_compile_serpent(code: str) -> spec.RpcRequest:
    return rpc_request.create('eth_compileSerpent', [code])




import pytest

from ctc import binary
from ctc import rpc


@pytest.mark.asyncio
async def test_web3_client_version():
    result = await rpc.async_web3_client_version()
    assert len(result) > 0


@pytest.mark.asyncio
async def test_web3_sha():
    data = '0x1234'
    result = await rpc.async_web3_sha3(data)
    keccak = binary.keccak(data, output_format='prefix_hex')
    assert keccak == result


@pytest.mark.asyncio
async def test_net_version():
    result = await rpc.async_net_version()
    assert len(result) > 0


@pytest.mark.asyncio
async def test_net_listening():
    listening = await rpc.async_net_listening()
    assert listening is not None


@pytest.mark.asyncio
async def test_eth_protocol_version():
    version = await rpc.async_eth_protocol_version()
    assert version is not None


@pytest.mark.asyncio
async def test_eth_syncing():
    syncing = await rpc.async_eth_syncing()
    assert syncing is not None




import typing

from ctc import binary
from ctc import spec
from . import contract_abi_io


async def async_get_event_abi(
    contract_abi: typing.Optional[spec.ContractABI] = None,
    contract_address: typing.Optional[spec.Address] = None,
    event_name: typing.Optional[str] = None,
    event_hash: typing.Optional[str] = None,
    event_abi: typing.Optional[spec.EventABI] = None,
    network: typing.Optional[spec.NetworkReference] = None,
):

    # get contract abi
    if contract_abi is None:
        contract_abi = await contract_abi_io.async_get_contract_abi(
            contract_address=contract_address,
            network=network,
        )

    return binary.get_event_abi(
        contract_abi=contract_abi,
        event_name=event_name,
        event_hash=event_hash,
        event_abi=event_abi,
    )




import toolstr
import tooltime

from ctc import binary
from ctc import evm
from ctc import rpc
from ctc import spec
from .. import block_utils


async def async_print_transaction_summary(transaction_hash, sort_logs_by=None):
    transaction = await rpc.async_eth_get_transaction_by_hash(transaction_hash)
    transaction_receipt = await rpc.async_eth_get_transaction_receipt(
        transaction_hash=transaction_hash
    )
    block = await block_utils.async_get_block(
        transaction['block_number'], include_full_transactions=False
    )

    from ctc.protocols import chainlink_utils

    eth_usd = await chainlink_utils.async_get_eth_price(
        block=transaction['block_number']
    )

    print('Transaction')
    print('-----------')
    print('- hash:', transaction['hash'])
    print('- from:', transaction['from'])
    print('- to:', transaction['to'])
    print()
    print()
    print('Receipt')
    print('-------')
    print('- success:', bool(transaction_receipt['status']))
    print('- time:', block['timestamp'])
    print('- timestamp:', tooltime.timestamp_to_iso(block['timestamp']))
    print('- block number:', transaction['block_number'])
    print(
        '- transaction index:',
        transaction_receipt['transaction_index'],
        '/',
        len(block['transactions']),
    )
    print(
        '- gas used:',
        toolstr.format(transaction_receipt['gas_used']),
        '/',
        toolstr.format(transaction['gas']),
    )
    print('- gas price:', toolstr.format(transaction['gas_price'] / 1e9))
    if 'max_priority_fee_per_gas' in transaction:
        print(
            '- priority + base:',
            toolstr.format(transaction['max_priority_fee_per_gas'] / 1e9),
            '+',
            toolstr.format(block['base_fee_per_gas'] / 1e9),
        )
    fee = transaction_receipt['gas_used'] * transaction['gas_price'] / 1e18
    fee_usd = fee * eth_usd
    print(
        '- total fee:',
        toolstr.format(fee),
        'ETH',
        '($' + toolstr.format(fee_usd) + ')',
    )

    if transaction['input'] == '0x':
        print()
        print()
        print('Call Data')
        print('---------')
        print('[none]')
    else:
        try:
            contract_abi = await evm.async_get_contract_abi(
                contract_address=transaction['to']
            )
        except spec.AbiNotFoundException:
            print()
            print()
            print('No Contract ABI Available')
            print('-------------------------')
            return

        function_abi = await evm.async_get_function_abi(
            contract_address=transaction['to'],
            function_selector=transaction['input'][:10],
        )
        call_data = binary.decode_call_data(
            call_data=transaction['input'],
            function_abi=function_abi,
        )
        print()
        print()
        print('Call Data')
        print('---------')
        print(
            call_data['function_selector'],
            '-->',
            binary.get_function_signature(function_abi=function_abi),
        )
        for p, (name, value) in enumerate(call_data['named_parameters'].items()):
            if (
                value
                == 115792089237316195423570985008687907853269984665640564039457584007913129639935
            ):
                value = 'INT_MAX'
            print('    ' + str(p + 1) + '.', name, '=', value)

    print()
    print()
    print('Logs')
    print('----')
    logs = transaction_receipt['logs']
    if len(logs) == 0:
        print('[none]')

    else:
        # if sort_logs_by == 'signature':
        #     logs = sorted(
        #         logs,
        #         key=lambda log: binary.get_event_signature(
        #             contract_address=log['address'],
        #             event_hash=log['topics'][0],
        #         ),
        #     )
        for l, log in enumerate(logs):
            if l != 0:
                print()

            event_abi = await evm.async_get_event_abi(
                contract_address=log['address'],
                event_hash=log['topics'][0],
            )
            normalized_event = binary.normalize_event(
                event=log,
                arg_prefix=None,
                event_abi=event_abi,
            )
            event_signature = binary.get_event_signature(event_abi=event_abi)
            print(event_signature, '-->', log['address'])
            for e, (name, value) in enumerate(normalized_event['args'].items()):
                if (
                    value
                    == 115792089237316195423570985008687907853269984665640564039457584007913129639935
                ):
                    value = 'INT_MAX'

                # not sure if this is what should be done
                if isinstance(value, bytes):
                    value = binary.convert(value, 'prefix_hex')

                print('    ' + str(e + 1) + '.', name, '=', value)




from __future__ import annotations

import typing

from ctc import spec


def get_binary_format(data: spec.BinaryInteger) -> spec.BinaryFormat:
    if isinstance(data, bytes):
        return 'binary'
    elif isinstance(data, str):
        if data.startswith('0x'):
            return 'prefix_hex'
        else:
            return 'raw_hex'
    elif isinstance(data, int):
        return 'integer'
    else:
        raise Exception('could not detect format')


def get_binary_n_bytes(data: spec.BinaryInteger) -> int:

    if isinstance(data, bytes):
        return len(data)
    elif isinstance(data, str):
        if len(data) % 2 != 0:
            raise Exception('hex data must have even number of characters')
        if data.startswith('0x'):
            return int(len(data) / 2) - 1
        else:
            return int(len(data) / 2)
    elif isinstance(data, int):
        # adapted from https://stackoverflow.com/a/30375198
        if data < 0:
            raise Exception('only positive integers allowed')
        return (data.bit_length() + 7) // 8
    else:
        raise Exception('unknown data type: ' + str(data))


#
# # binary data manipulation
#


@typing.overload
def convert(
    data: spec.BinaryInteger,
    output_format: typing.Literal['binary'],
) -> bytes:
    ...


@typing.overload
def convert(
    data: spec.BinaryInteger,
    output_format: typing.Literal['integer'],
) -> int:
    ...


@typing.overload
def convert(
    data: spec.BinaryInteger,
    output_format: typing.Optional[typing.Literal['prefix_hex', 'raw_hex']],
) -> str:
    ...


def convert(
    data: spec.BinaryInteger,
    output_format: typing.Optional[spec.BinaryFormat] = None,
) -> spec.BinaryInteger:
    """convert {hex str or bytes} into {hex str or bytes}

    function should not be used with general text data

    ## Data Types
    - 'prefix_hex': hex str with 0x prefix included
    - 'raw_hex': hex str without 0x prefix included
    - 'binary': bytes
    """

    if output_format is None:
        output_format = 'prefix_hex'

    if isinstance(data, str):
        if data.startswith('0x'):
            raw_data = data[2:]
        else:
            raw_data = data

        if output_format == 'prefix_hex':
            return '0x' + raw_data
        elif output_format == 'raw_hex':
            return raw_data
        elif output_format == 'binary':
            return bytes.fromhex(raw_data)
        elif output_format == 'integer':
            return int(data, 16)
        else:
            raise Exception('invalid output_format: ' + str(output_format))

    elif isinstance(data, bytes):

        if output_format == 'binary':
            return data
        elif output_format == 'prefix_hex':
            return '0x' + data.hex()
        elif output_format == 'raw_hex':
            return data.hex()
        elif output_format == 'integer':
            return int.from_bytes(data, 'big')
        else:
            raise Exception('invalid output_format: ' + str(output_format))

    elif isinstance(data, int):

        if data < 0:
            raise Exception('only positive integers allowed')

        n_bytes = get_binary_n_bytes(data)

        if output_format == 'binary':
            return data.to_bytes(n_bytes, 'big')
        elif output_format == 'prefix_hex':
            return hex(data)
        elif output_format == 'raw_hex':
            return hex(data)[2:]
        elif output_format == 'integer':
            return data
        else:
            raise Exception('invalid output_format: ' + str(output_format))

    else:

        raise Exception('unknown input data format: ' + str(type(data)))


def add_binary_pad(
    data: spec.BinaryInteger,
    pad_side: typing.Literal['left', 'right', None] = None,
    padded_size: int = None,
) -> spec.BinaryInteger:
    """add pad of zeros to left or right side of binary data"""

    # default arguments
    if pad_side is None:
        pad_side = 'left'
    if padded_size is None:
        padded_size = 32

    # determine pad bytes
    binary_format = get_binary_format(data)
    data_bytes = get_binary_n_bytes(data)
    if padded_size < data_bytes:
        raise Exception('pad size too small for data')
    pad_bytes = padded_size - data_bytes

    if binary_format == 'binary':

        data = typing.cast(bytes, data)

        if pad_side == 'left':
            return bytes(0) * pad_bytes + data
        elif pad_side == 'right':
            return data + bytes(0) * pad_bytes
        else:
            raise Exception('invalid pad side: ' + str(pad_side))

    elif binary_format == 'prefix_hex':

        data = typing.cast(str, data)

        if pad_side == 'left':
            return '0x' + '0' * 2 * pad_bytes + data[2:]
        elif pad_side == 'right':
            return data + '0' * 2 * pad_bytes
        else:
            raise Exception('invalid pad side: ' + str(pad_side))

    elif binary_format == 'raw_hex':

        data = typing.cast(str, data)

        if pad_side == 'left':
            return '0' * 2 * pad_bytes + data
        elif pad_side == 'right':
            return data + '0' * 2 * pad_bytes
        else:
            raise Exception('invalid pad side: ' + str(pad_side))

    else:

        raise Exception('invalid binary format: ' + str(binary_format))


def match_format(
    format_this: spec.BinaryInteger,
    like_this: spec.BinaryInteger,
    match_pad: bool = False,
) -> spec.BinaryInteger:
    """

    will only match left pads, because pad size cannot be reliably determined
    """

    output_format = get_binary_format(like_this)
    output = convert(data=format_this, output_format=output_format)

    if match_pad:
        padded_size = get_binary_n_bytes(like_this)
        output = add_binary_pad(output, padded_size=padded_size)

    return output




from ctc import spec


def digest_eth_get_work(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_submit_work(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_submit_hashrate(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_coinbase(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_mining(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_hashrate(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response




from __future__ import annotations

import typing


class NoMatchFound(LookupError):
    pass


class MultipleMatchesFound(LookupError):
    pass


M = typing.TypeVar('M', bound=typing.Mapping)


def get_matching_entries(
    sequence: typing.Sequence[M],
    query: typing.Mapping[typing.Any, typing.Any],
) -> list[M]:
    matches: list[M] = []
    for item in sequence:
        for key, value in query.items():
            if item.get(key) != value:
                break
        else:
            matches.append(item)
    return matches


@typing.overload
def get_matching_entry(
    sequence: typing.Sequence[M],
    query: typing.Mapping[typing.Any, typing.Any],
    raise_if_not_found: typing.Literal[False],
) -> typing.Optional[M]:
    ...


@typing.overload
def get_matching_entry(
    sequence: typing.Sequence[M],
    query: typing.Mapping[typing.Any, typing.Any],
    raise_if_not_found: typing.Literal[True] = True,
) -> M:
    ...


def get_matching_entry(
    sequence: typing.Sequence[M],
    query: typing.Mapping[typing.Any, typing.Any],
    raise_if_not_found: bool = True,
) -> typing.Optional[M]:
    matches = get_matching_entries(sequence=sequence, query=query)
    if len(matches) == 1:
        return matches[0]
    elif len(matches) > 1:
        raise MultipleMatchesFound('more than one match found in sequence')
    else:
        if raise_if_not_found:
            raise NoMatchFound('no matches found in sequence')
        else:
            return None


def binary_search(
    *,
    is_match,
    start_index=None,
    end_index=None,
    index_range=None,
    raise_if_not_found=True
):
    """return the first index for which match returns True"""

    if start_index is None and end_index is None:
        start_index, end_index = index_range

    start_index = int(start_index)
    end_index = int(end_index)

    if is_match(start_index):
        return start_index
    if not is_match(end_index):
        if raise_if_not_found:
            raise NoMatchFound('could not find match')
        else:
            return None

    while True:

        midpoint = (start_index + end_index) / 2
        midpoint = int(midpoint)

        if is_match(midpoint):
            end_index = midpoint
        else:
            start_index = midpoint

        if start_index + 1 == end_index:
            return end_index


async def async_binary_search(
    *,
    async_is_match,
    start_index=None,
    end_index=None,
    index_range=None,
    raise_if_not_found=True
):
    """return the first index for which match returns True"""

    if start_index is None and end_index is None:
        start_index, end_index = index_range

    start_index = int(start_index)
    end_index = int(end_index)

    if await async_is_match(start_index):
        return start_index
    if not await async_is_match(end_index):
        if raise_if_not_found:
            raise NoMatchFound('could not find match')
        else:
            return None

    while True:

        midpoint = (start_index + end_index) / 2
        midpoint = int(midpoint)

        if await async_is_match(midpoint):
            end_index = midpoint
        else:
            start_index = midpoint

        if start_index + 1 == end_index:
            return end_index


def nary_search(
    nary,
    start_index,
    end_index,
    is_match,
    debug=False,
    raise_if_not_found=True,
    get_next_probes=None,
):

    if get_next_probes is None:
        get_next_probes = get_next_probes_linear

    extra_probes = [start_index, end_index]
    probe_min = start_index
    probe_max = end_index

    while True:

        # if probe range minimized, return result
        if probe_max == probe_min + 1:
            return probe_max

        # get next probes to test
        probes = get_next_probes(
            probe_min=probe_min, probe_max=probe_max, nary=nary
        )
        probes = sorted(set(probes))
        n_probes = len(probes)

        # add in extra probes for start_index and end_index
        all_probes = probes + extra_probes

        # compute results
        all_results = is_match(all_probes)
        results = all_results[:n_probes]
        extra_results = all_results[n_probes:]

        # separate start_index and end_index probes
        if len(extra_probes) > 0:
            start_result, end_result = extra_results
            if start_result:
                return start_index
            elif not end_result:
                if raise_if_not_found:
                    raise Exception('search range does not go high enough')
                else:
                    return None
            extra_probes = []

        # determine lowest successful probe
        for p in range(len(probes)):
            if results[p]:
                break
        else:
            p += 1

        # print state
        if debug:
            print('probe_min:', probe_min)
            print('probe_max:', probe_max)
            print('n_probes:', n_probes)
            print('probes:', probes)
            print('results:', results)
            print('p:', p)
            print()

        # adjust search boundaries
        if p == 0:
            probe_min = probe_min
            probe_max = probes[0]
        elif p == len(probes):
            probe_min = probes[-1]
            probe_max = probe_max
        else:
            probe_min = probes[p - 1]
            probe_max = probes[p]


async def async_nary_search(
    nary,
    start_index,
    end_index,
    async_is_match,
    debug=False,
    raise_if_not_found=True,
    get_next_probes=None,
):

    if get_next_probes is None:
        get_next_probes = get_next_probes_linear

    extra_probes = [start_index, end_index]
    probe_min = start_index
    probe_max = end_index

    while True:

        # if probe range minimized, return result
        if probe_max == probe_min + 1:
            return probe_max

        # get next probes to test
        probes = get_next_probes(
            probe_min=probe_min, probe_max=probe_max, nary=nary
        )
        probes = sorted(set(probes))
        n_probes = len(probes)

        # add in extra probes for start_index and end_index
        all_probes = probes + extra_probes

        # compute results
        all_results = await async_is_match(all_probes)
        results = all_results[:n_probes]
        extra_results = all_results[n_probes:]

        # separate start_index and end_index probes
        if len(extra_probes) > 0:
            start_result, end_result = extra_results
            if start_result:
                return start_index
            elif not end_result:
                if raise_if_not_found:
                    raise Exception('search range does not go high enough')
                else:
                    return None
            extra_probes = []

        # determine lowest successful probe
        for p in range(len(probes)):
            if results[p]:
                break
        else:
            p += 1

        # print state
        if debug:
            print('probe_min:', probe_min)
            print('probe_max:', probe_max)
            print('n_probes:', n_probes)
            print('probes:', probes)
            print('results:', results)
            print('p:', p)
            print()

        # adjust search boundaries
        if p == 0:
            probe_min = probe_min
            probe_max = probes[0]
        elif p == len(probes):
            probe_min = probes[-1]
            probe_max = probe_max
        else:
            probe_min = probes[p - 1]
            probe_max = probes[p]


def get_next_probes_linear(probe_min, probe_max, nary):
    n_probes = min(nary - 1, probe_max - probe_min - 1)
    d = (probe_max - probe_min) / (n_probes + 1)
    probes = [probe_min + (p + 1) * d for p in range(n_probes)]
    probes = [round(probe) for probe in probes]
    return probes




from ctc import rpc

from .. import rari_abis


fuse_directory = '0x835482fe0532f169024d5e9410199369aad5c77e'


async def async_get_all_pools(block='latest'):
    # TODO: convert output to dict
    return await rpc.async_eth_call(
        to_address=fuse_directory,
        block_number=block,
        function_abi=rari_abis.pool_directory_abis['getAllPools'],
    )





contenttypes = {
    'AMMPool': {
        'x_reserves': 'Number',
        'y_reserves': 'Number',
        'lp_total_supply': 'Number',
    },
    'AMMTrade': {
        'x_bought': 'Number',
        'x_sold': 'Number',
        'y_bought': 'Number',
        'y_sold': 'Number',
        'fee_rate': 'Number',
        'new_pool': 'AMMPool',
    },
    'AMMMint': {
        'x_deposited': 'Number',
        'y_deposited': 'Number',
        'lp_minted': 'Number',
        'new_pool': 'AMMPool',
    },
    'AMMBurn': {
        'x_withdrawn': 'Number',
        'y_withdrawn': 'Number',
        'lp_burned': 'Number',
        'new_pool': 'AMMPool',
    },
    'AMMPoolSummary': {
        'x_per_y': 'Number',
        'y_per_x': 'Number',
        'x_per_y_depth': {'Number': 'Number'},
        'y_per_x_depth': {'Number': 'Number'},
    },
    'AMMTradeSummary': {
        'slippage': 'Number',
        'mean_x_per_y': 'Number',
        'mean_y_per_x': 'Number',
        'x_fees': 'Number',
        'y_fees': 'Number',
    },
}




from ctc import binary
from ctc import rpc
from ctc import spec

from .. import abi_utils
from . import address_data


async def async_print_address_summary(
    address, verbose=0, max_width=80, provider: spec.ProviderSpec = None
):
    """

    TODO (with more data)
    - transactions
    - transfers
    """

    eth_balance = await rpc.async_eth_get_balance(address, provider=provider)
    transaction_count = await rpc.async_eth_get_transaction_count(
        address, provider=provider
    )

    # contract data
    code = await rpc.async_eth_get_code(address, provider=provider)
    is_contract = code != '0x'
    if is_contract:
        address_type = 'contract'
    else:
        address_type = 'EOA'

    title = 'Address ' + address.lower()
    print(title)
    print('─' * len(title))
    print('- checksum:', address_data.get_address_checksum(address))
    print('- address type:', address_type)
    print('- ETH balance:', eth_balance)
    print('- transaction count:', transaction_count)
    print('- first transaction:')

    if is_contract:
        print()
        print()

        provider = rpc.get_provider(provider)
        network = provider['network']
        contract_abi = await abi_utils.async_get_contract_abi(
            contract_address=address, network=network,
        )
        df = abi_utils.contract_abi_to_dataframe(
            contract_abi=contract_abi, human_readable=False
        )

        functions = df[df['type'] == 'function']
        functions = functions.drop(columns=['type', 'anonymous'])

        print('Contract ABI Functions')
        print('──────────────────────')
        for i, (f, function) in enumerate(functions.iterrows()):

            if len(function['outputs']) == 0:
                output_str = '[none]'
            else:
                output_str_list = [
                    item['type'] + ' ' + item['name']
                    for item in function['outputs']
                ]
                output_str = ', '.join(output_str_list)
                if len(function['outputs']) > 1:
                    output_str = '(' + output_str + ')'

            if not verbose:
                signature = binary.get_function_signature(
                    function_abi=function, include_names=True
                )
            else:
                signature = function['name'] + '()'
            if verbose:
                text = signature + ' --> ' + str(output_str)
            else:
                text = str(i + 1) + '. ' + signature + ' --> ' + str(output_str)

            if len(text) > max_width:
                text = text[: max_width - 3] + '...'

            print(text)
            if verbose:
                indent = ''
                if verbose > 1:
                    indent = '    '
                    print(indent + '- mutability:', function['stateMutability'])
                    print(indent + '- inputs:')
                if len(function['inputs']) == 0:
                    print(indent + '    [no inputs]')
                for i, item in enumerate(function['inputs']):
                    print(
                        indent + '    ' + str(i + 1) + '.',
                        item['type'],
                        item['name'],
                    )
                print()

        events = df[df['type'] == 'event']
        print()
        print('Contract ABI Events')
        print('───────────────────')
        if len(events) == 0:
            print('[none]')
        for i, (e, event) in enumerate(events.iterrows()):
            event_hash = binary.get_event_hash(event_abi=event)
            signature = binary.get_event_signature(event_abi=event)
            line = str(i + 1) + '. ' + signature
            if len(line) > max_width:
                line = line[: max_width - 3] + '...'
            print(line)
            if verbose:
                print('  ', event_hash)
                if i + 1 != len(events):
                    print()




from __future__ import annotations

import asyncio
import typing

from ctc import evm
from ctc import rpc
from ctc import spec
from . import coracle_spec
from . import coracle_tokens


async def async_get_tokens_deposits(
    tokens: typing.Optional[typing.Sequence[spec.Address]] = None,
    block: spec.BlockReference = 'latest',
    provider: spec.ProviderSpec = None,
) -> dict[spec.Address, typing.Tuple[spec.ContractAddress, ...]]:
    """get all deposits of all tokens in pcv"""

    block = await evm.async_block_number_to_int(block=block, provider=provider)

    # get tokens in pcv
    if tokens is None:
        tokens = await coracle_tokens.async_get_tokens_in_pcv(
            block=block, provider=provider
        )

    # get deposits of each token
    coroutines = []
    for token in tokens:
        coroutine = async_get_token_deposits(
            token=token,
            block=block,
            provider=provider,
        )
        coroutines.append(coroutine)

    # compile into tokens_deposits
    results = await asyncio.gather(*coroutines)
    tokens_deposits = dict(zip(tokens, results))
    return tokens_deposits


async def async_get_token_deposits(
    token: spec.Address,
    block: spec.BlockNumberReference = 'latest',
    wrapper: bool = False,
    provider: spec.ProviderSpec = None,
) -> typing.Tuple[spec.ContractAddress, ...]:
    """get list of a token's deposits"""

    block = await evm.async_block_number_to_int(block=block, provider=provider)
    coracle = coracle_spec.get_coracle_address(wrapper=wrapper, block=block)
    return await rpc.async_eth_call(
        to_address=coracle,
        block_number=block,
        function_name='getDepositsForToken',
        function_parameters={'_token': token},
        provider=provider,
    )


async def async_get_deposit_token(
    deposit: spec.ContractAddress,
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
) -> spec.Address:
    """get the token address of a deposit"""
    return await rpc.async_eth_call(
        to_address=deposit,
        block_number=block,
        function_name='token',
        provider=provider,
    )




"""
# balance of single wallet across multiple tokens;
ctc erc20 balances <wallet> --erc20s <erc20s> [--block <block>]

# balances of wallets within single block (default = all wallets)
ctc erc20 balances <erc20> [--block <block>] [--wallets <wallets>]

# balance of single wallet across multiple blocks
ctc erc20 balances <erc20> <wallet> --blocks <blocks>
"""

import pandas as pd

import toolstr

from ctc import evm
from ctc import rpc
from ctc.cli import cli_utils


def get_command_spec():
    return {
        'f': async_balances_command,
        'args': [
            {'name': 'args', 'kwargs': {'nargs': '+'}},
            {'name': '--block'},
            {'name': '--wallets', 'kwargs': {'nargs': '+'}},
            {'name': '--blocks', 'kwargs': {'nargs': '+'}},
            {'name': '--erc20s', 'kwargs': {'nargs': '+'}},
            #
            {'name': '--raw', 'kwargs': {'action': 'store_true'}},
            {'name': '--output', 'kwargs': {'default': 'stdout'}},
            {'name': '--overwrite', 'kwargs': {'action': 'store_true'}},
            {'name': '--top'},
        ],
    }


async def async_balances_command(
    args, block, wallets, blocks, erc20s, raw, output, overwrite, top
):
    indent = None

    if len(args) == 1 and erc20s is not None:
        # multiple erc20s, single wallet, single block
        if wallets is not None or blocks is not None:
            raise Exception(
                'can only specify one of --erc20s --wallets, --blocks'
            )
        wallet = args[0]

        if block is None:
            block = 'latest'
        block = await evm.async_block_number_to_int(block)
        symbols_coroutine = evm.async_get_erc20s_symbols(erc20s)
        balances = await evm.async_get_erc20s_balance_of(
            address=wallet,
            tokens=erc20s,
            block=block,
            normalize=(not raw),
        )
        symbols = await symbols_coroutine
        data = {'balance': balances, 'symbol': symbols, 'erc20_address': erc20s}
        df = pd.DataFrame(data)
        df = df.set_index('erc20_address')

        toolstr.print_text_box('ERC20 balances in wallet')
        print('- wallet:', wallet)
        print('- block:', block)
        print('- n_tokens:', len(erc20s))
        print()
        print()
        indent = '    '

    elif len(args) == 1:
        # single erc20, multiple wallets, single block
        if blocks is not None or erc20s is not None:
            raise Exception(
                'can only specify one of --erc20s --wallets, --blocks'
            )
        erc20 = args[0]

        if block is None:
            block = 'latest'
        block = await evm.async_block_number_to_int(block)
        symbol_coroutine = evm.async_get_erc20_symbol(erc20)

        if wallets is not None:
            balances = await evm.async_get_erc20_balance_of_addresses(
                addresses=wallets,
                token=erc20,
                block=block,
                normalize=(not raw),
            )
            symbol = await symbol_coroutine

            df = pd.Series(balances, index=wallets)
            df.name = 'balance'
            df.index.name = 'address'

            print()
            toolstr.print_text_box(symbol + ' Balances')
            print('- token:', erc20)
            print('- symbol:', symbol)
            print('- block:', block)
            print()
            print()

        else:
            transfers = await evm.async_get_erc20_transfers(
                erc20,
                start_block=None,
                end_block=block,
                normalize=False,
            )
            df = await evm.async_get_erc20_holdings_from_transfers(
                transfers=transfers, dtype=None, normalize=(not raw)
            )
            symbol = await symbol_coroutine

            if top is None:
                top = 20

            print()
            toolstr.print_text_box(symbol + ' Balances')
            print('- token:', erc20)
            print('- symbol:', symbol)
            print('- block:', block)
            print()
            print()
            toolstr.print_header('Showing Top ' + str(top) + ' Holders')
            print()
            indent = '    '

    elif len(args) == 2 and blocks is not None:
        # single erc20, single address, multiple blocks
        if erc20s is not None or wallets is not None:
            raise Exception(
                'can only specify one of --erc20s --wallets, --blocks'
            )
        erc20, wallet = args

        resolved_blocks = await cli_utils.async_resolve_blocks(blocks)
        balances = await evm.async_get_erc20_balance_of_by_block(
            address=wallet,
            token=erc20,
            blocks=resolved_blocks,
            normalize=(not raw),
            empty_token=0,
        )
        df = pd.DataFrame(balances, index=resolved_blocks)
        df.index.name = 'block'
        df.columns = ['balance']

    else:
        raise Exception('invalid inputs')

    cli_utils.output_data(df, output, overwrite, top=top, indent=indent)

    await rpc.async_close_http_session()




from __future__ import annotations

import asyncio
import typing

from ctc import directory
from ctc import evm
from ctc import rpc
from ctc import spec

from . import pool_metadata


#
# # weights
#


async def async_get_pool_weights_raw(
    pool_address: spec.ContractAddress,
    block: spec.BlockNumberReference = 'latest',
) -> typing.Union[list[int], list[float]]:

    return await rpc.async_eth_call(
        to_address=pool_address,
        function_name='getNormalizedWeights',
        block_number=block,
    )


async def async_get_pool_weights(
    pool_address: spec.ContractAddress,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
) -> typing.Union[
    dict[spec.ContractAddress, int],
    dict[spec.ContractAddress, float],
]:

    tokens_coroutine = pool_metadata.async_get_pool_tokens(
        pool_address=pool_address, block=block
    )
    weights_coroutine = async_get_pool_weights_raw(
        pool_address=pool_address, block=block
    )

    tokens, weights = await asyncio.gather(tokens_coroutine, weights_coroutine)

    if normalize:
        weights = [weight / 1e18 for weight in weights]

    return dict(zip(tokens, weights))


async def async_get_pool_weights_by_block(
    pool_address: spec.ContractAddress,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
) -> typing.Union[
    dict[spec.BlockNumberReference, int],
    dict[spec.BlockNumberReference, float],
]:

    weights = await rpc.async_batch_eth_call(
        to_address=pool_address,
        function_name='getNormalizedWeights',
        block_numbers=blocks,
        provider={'chunk_size': 100},
    )

    if normalize:
        weights = [
            [block_weight / 1e18 for block_weight in block_weights]
            for block_weights in weights
        ]

    return dict(zip(blocks, weights))


#
# # fees
#


async def async_get_pool_fees(
    pool_address: spec.ContractAddress,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
):

    fees = await rpc.async_eth_call(
        to_address=pool_address,
        function_name='getSwapFeePercentage',
        block_number=block,
    )

    if normalize:
        fees /= 1e18

    return fees


#
# # balances
#


async def async_get_pool_balances(
    *,
    pool_address: typing.Optional[spec.ContractAddress] = None,
    pool_id: typing.Optional[spec.HexData] = None,
    block: spec.BlockNumberReference = 'latest',
    vault: typing.Optional[spec.ContractAddress] = None,
    normalize: bool = True,
) -> typing.Union[
    dict[spec.Address, int],
    dict[spec.Address, float],
]:

    if vault is None:
        vault = directory.get_address(name='Vault', label='balancer')
    if pool_id is None:
        pool_id = await pool_metadata.async_get_pool_id(
            pool_address, block=block
        )

    pool_tokens = await rpc.async_eth_call(
        to_address=vault,
        function_name='getPoolTokens',
        function_parameters=[pool_id],
        block_number=block,
        package_named_outputs=True,
    )

    pool_balances = dict(zip(pool_tokens['tokens'], pool_tokens['balances']))

    if normalize:
        tokens = pool_balances.keys()
        decimals = await evm.async_get_erc20s_decimals(
            tokens=tokens,
            block=block,
        )
        for token, decimal in zip(tokens, decimals):
            pool_balances[token] /= 10 ** decimal

    return pool_balances




import asyncio

from ctc import evm
from ctc import rpc


def get_command_spec():
    return {
        'f': add_proxy_abi_comamand,
        'args': [
            {'name': 'contract_address'},
            {'name': 'implementation_address'},
        ],
    }


def add_proxy_abi_comamand(contract_address, implementation_address, **kwargs):
    print('saving proxy contract implementation...')
    print('     for contract:', contract_address)
    print('   implementation:', implementation_address)
    answer = input('continue? ')
    if answer not in ['y', 'yes']:
        print('aborting')
        return
    else:
        asyncio.run(run(contract_address, implementation_address))


async def run(contract_address, implementation_address):
    await evm.async_save_proxy_contract_abi_to_filesystem(
        contract_address=contract_address,
        proxy_implementation=implementation_address,
    )

    from ctc.rpc.rpc_backends import rpc_http_async

    provider = rpc.get_provider()
    await rpc_http_async.async_close_session(provider=provider)




import numpy as np
import pandas as pd

import tooltime

from ctc import evm
from ctc import rpc
from ctc import spec

from ctc.cli import cli_utils


def get_command_spec():
    return {
        'f': async_block_command,
        'args': [
            {'name': 'blocks', 'kwargs': {'nargs': '+'}},
            {'name': '--attributes', 'kwargs': {'nargs': '+'}},
            {'name': '--output', 'kwargs': {'default': 'stdout'}},
            {'name': '--overwrite', 'kwargs': {'action': 'store_true'}},
            {'name': '--provider', 'kwargs': {}},
        ],
    }


async def async_block_command(
    attributes,
    blocks,
    output,
    provider,
    overwrite,
):

    if attributes is None:
        # gas stats as well
        attributes = [
            'number',
            'timestamp',
            'time',
            'min_gas_price',
            'median_gas_price',
            'mean_gas_price',
            'max_gas_price',
        ]
    elif attributes == ['all']:
        attributes = list(spec.Block.__annotations__.keys())
    else:
        attributes = [
            attribute for token in attributes for attribute in token.split(',')
        ]

    # determine blocks
    export_blocks = await cli_utils.async_resolve_blocks(blocks)

    # print summary
    print('exporting from', len(export_blocks), 'blocks')
    print('- min:', min(export_blocks))
    print('- max:', max(export_blocks))
    print('- attributes:', attributes)
    print()

    # obtain data
    gas_attrs = [
        'min_gas_price',
        'median_gas_price',
        'mean_gas_price',
        'max_gas_price',
    ]
    compute_gas = any(attr in attributes for attr in gas_attrs)
    blocks_data = await rpc.async_batch_eth_get_block_by_number(
        block_numbers=export_blocks,
        include_full_transactions=compute_gas,
        overwrite=overwrite,
        provider=provider,
    )

    # format as dataframe
    df = pd.DataFrame(blocks_data)

    # special attribute: gas stats
    if compute_gas:
        gas_stats = pd.DataFrame(
            evm.get_block_gas_stats(block_data)
            for block_data in blocks_data
        )
        for gas_attr in gas_attrs:
            if gas_attr in attributes:
                df[gas_attr] = gas_stats[gas_attr]

    # special attribute: base_fee_per_gas
    if 'base_fee_per_gas' in attributes and 'base_fee_per_gas' not in df:
        df['base_fee_per_gas'] = np.nan
    if 'base_fee_per_gas' in df:
        df['base_fee_per_gas'] /= 1e9

    # special attribute: time
    if 'time' in attributes:
        df['time'] = df['timestamp'].map(tooltime.timestamp_to_iso)

    # trim attributes
    if len(attributes) > 0:
        df = df[attributes]

    # output data
    if 'number' in df:
        df = df.set_index('number')
    cli_utils.output_data(df=df, output=output, overwrite=overwrite)

    # cleanup
    await rpc.async_close_http_session()




import asyncio

import pytest


@pytest.fixture(scope="session")
def event_loop(request):
    """create an instance of the default event loop for each test case

    adapted from https://stackoverflow.com/a/66225169
    """
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()




from __future__ import annotations

import asyncio
import typing

from ctc import spec
from . import block_crud


#
# # conversions
#


async def async_block_number_to_int(
    block: spec.BlockNumberReference,
    provider: spec.ProviderSpec = None,
) -> int:
    """resolve block number reference to int (e.g. converting 'latest' to int)

    Examples: 'latest', or 9999.0, or 9999
    """
    if block in spec.block_number_names:
        return await block_crud.async_get_latest_block_number(provider=provider)
    else:
        return raw_block_number_to_int(block)


def standardize_block_number(
    block: spec.BlockNumberReference,
) -> spec.StandardBlockNumber:
    """turn block into standard block number reference

    Examples: 'latest' or 123456
    """

    if block in spec.block_number_names:
        return typing.cast(spec.BlockNumberName, block)
    else:
        return raw_block_number_to_int(block)


def raw_block_number_to_int(block: spec.RawBlockNumber) -> int:
    """convert block number to int"""

    # python3.7 compatibility
    # supports_int = isinstance(block, typing.SupportsInt)
    supports_int = hasattr(block, '__int__')

    if supports_int:
        if isinstance(block, int):
            return block
        else:
            block_supports_int = typing.cast(typing.SupportsInt, block)
            as_int = int(round(block_supports_int))  # type: ignore
            if abs(as_int - int(block_supports_int)) > 0.0001:
                raise Exception('must specify integer blocks')
            return as_int
    elif isinstance(block, str):
        if block.startswith('0x'):
            try:
                return int(block, 16)
            except ValueError:
                pass
        elif 'e' in block:
            as_float = float(block)
            as_int = int(as_float)
            if abs(as_float - as_int) > 0.0001:
                raise Exception('must specify integer blocks')
            else:
                return as_int
        else:
            return int(block)

    raise Exception('unknown block number specification: ' + str(block))


#
# # plural versions
#


async def async_block_numbers_to_int(
    blocks: typing.Iterable[spec.BlockNumberReference],
    provider: spec.ProviderSpec = None,
) -> list[int]:
    coroutines = [
        async_block_number_to_int(block=block, provider=provider)
        for block in blocks
    ]
    return await asyncio.gather(*coroutines)


def standardize_block_numbers(
    blocks: typing.Iterable[spec.BlockNumberReference],
) -> list[spec.StandardBlockNumber]:
    """standardize an iterable of block number references"""
    return [standardize_block_number(block) for block in blocks]


def raw_block_number_to_ints(
    blocks: typing.Iterable[spec.RawBlockNumber],
) -> list[int]:
    return [raw_block_number_to_int(block) for block in blocks]




from __future__ import annotations

from ctc import rpc


async def async_get_block_gas_stats(block, normalize=True):
    if not isinstance(block, dict):
        if isinstance(block, str) and block.startswith('0x') and len(block) == 66:
            block = await rpc.async_eth_get_block_by_hash(block)
        else:
            block = await rpc.async_eth_get_block_by_number(block)
    return get_block_gas_stats(block)


def get_block_gas_stats(block, normalize=True):
    import numpy as np

    base_fee = block.get('base_fee_per_gas')

    if len(block['transactions']) > 0:
        if isinstance(block['transactions'][0], str):
            raise Exception('transaction data not in block')

        gas_prices = [
            transaction['gas_price'] for transaction in block['transactions']
        ]

        if normalize:
            gas_prices = [gas_price / 1e9 for gas_price in gas_prices]
            if base_fee is not None:
                base_fee /= 1e9

        min_gas_price = min(gas_prices)
        median_gas_price = np.median(gas_prices)
        mean_gas_price = sum(gas_prices) / len(gas_prices)
        max_gas_price = max(gas_prices)

    else:
        min_gas_price = None
        median_gas_price = None
        mean_gas_price = None
        max_gas_price = None

    return {
        'base_fee': base_fee,
        'min_gas_price': min_gas_price,
        'median_gas_price': median_gas_price,
        'mean_gas_price': mean_gas_price,
        'max_gas_price': max_gas_price,
        'gas_limit': block['gas_limit'],
        'gas_used': block['gas_used'],
        'n_transactions': len(block['transactions']),
    }




import asyncio
import typing

from ctc import evm
from ctc import spec
from ctc.toolbox import nested_utils

from . import uniswap_v2_metadata
from . import uniswap_v2_spec


async def replace_pool_state_symbols(
    pool_state: typing.Mapping,
    *,
    pool: spec.Address = None,
    symbols: typing.Optional[typing.Sequence[str]] = None,
    provider: spec.ProviderSpec = None,
) -> typing.Mapping:

    if symbols is None:
        if pool is None:
            raise Exception('must specify pool or symbols')
        symbols = await uniswap_v2_metadata.async_get_pool_symbols(
            pool=pool, provider=provider
        )

    x_symbol, y_symbol = symbols

    return {
        x_symbol + '_reserves': pool_state['x_reserves'],
        y_symbol + '_reserves': pool_state['y_reserves'],
        'lp_total_supply': pool_state['lp_total_supply'],
    }


async def async_get_pool_state(
    pool: spec.Address,
    *,
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
) -> uniswap_v2_spec.PoolState:

    block = await evm.async_block_number_to_int(block, provider=provider)

    # reserves
    token_x, token_y = await uniswap_v2_metadata.async_get_pool_tokens(
        pool=pool, provider=provider
    )
    reserves_coroutine = evm.async_get_erc20s_balance_of(
        address=pool,
        tokens=[token_x, token_y],
        block=block,
        provider=provider,
        normalize=normalize,
    )
    reserves_task = asyncio.create_task(reserves_coroutine)

    # total supply
    lp_total_supply_coroutine = evm.async_get_erc20_total_supply(
        token=pool,
        block=block,
        provider=provider,
        normalize=normalize,
    )
    lp_total_supply_task = asyncio.create_task(lp_total_supply_coroutine)

    # metadata
    if normalize:
        decimals_coroutine = uniswap_v2_metadata.async_get_pool_decimals(
            pool=pool,
            provider=provider,
        )
        decimals_task = asyncio.create_task(decimals_coroutine)

    # await results
    token_x_reserves, token_y_reserves = await reserves_task
    lp_total_supply = await lp_total_supply_task

    output: uniswap_v2_spec.PoolState = {
        'x_reserves': token_x_reserves,
        'y_reserves': token_y_reserves,
        'lp_total_supply': lp_total_supply,
    }
    return output


async def async_get_pool_state_by_block(
    pool: spec.Address,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
) -> uniswap_v2_spec.PoolStateByBlock:

    if normalize:
        decimals_coroutine = uniswap_v2_metadata.async_get_pool_decimals(
            pool=pool,
            provider=provider,
        )
        decimals_task = asyncio.create_task(decimals_coroutine)

    coroutines = [
        async_get_pool_state(
            pool=pool,
            block=block,
            provider=provider,
            normalize=False,
        )
        for block in blocks
    ]
    results: list[uniswap_v2_spec.PoolState] = await asyncio.gather(*coroutines)

    output: uniswap_v2_spec.PoolStateByBlock = (
        nested_utils.list_of_dicts_to_dict_of_lists(results)  # type: ignore
    )

    if normalize:

        x_decimals, y_decimals = await decimals_task

        output['x_reserves'] = [
            item / (x_decimals ** 18) for item in output['x_reserves']
        ]
        output['y_reserves'] = [
            item / (y_decimals ** 18) for item in output['y_reserves']
        ]
        output['lp_total_supply'] = [
            item / (10 ** 18) for item in output['lp_total_supply']
        ]

    return output




from __future__ import annotations

import asyncio
import logging
import math
import random
import typing

from ctc import spec
from . import rpc_provider


def create(method: str, parameters: list[typing.Any]) -> spec.RpcRequest:
    return {
        'jsonrpc': '2.0',
        'method': method,
        'params': parameters,
        'id': random.randint(1, int(1e18)),
    }


def send(
    request: spec.RpcRequest,
    provider: typing.Optional[spec.ProviderSpec] = None,
) -> spec.RpcResponse:
    full_provider = rpc_provider.get_provider(provider)

    if isinstance(request, dict):
        response = send_raw(request=request, provider=full_provider)
        return response['result']

    elif isinstance(request, list):

        # chunk request
        request_chunks = chunk_request(request=request, provider=full_provider)

        # send request chunks
        response_chunks = [
            send_raw(request=request_chunk, provider=full_provider)
            for request_chunk in request_chunks
        ]

        # reorder chunks
        plural_response = reorder_response_chunks(response_chunks, request)

        return [subresponse['result'] for subresponse in plural_response]

    else:

        raise Exception('unknown request type: ' + str(type(request)))


@typing.overload
def send_raw(
    request: spec.RpcSingularRequest, provider
) -> spec.RpcSingularResponseRaw:
    ...


@typing.overload
def send_raw(
    request: spec.RpcPluralRequest, provider
) -> spec.RpcPluralResponseRaw:
    ...


def send_raw(request: spec.RpcRequest, provider) -> spec.RpcResponseRaw:

    if provider['protocol'] == 'http':
        from .rpc_backends import rpc_http

        return rpc_http.send_http(
            request=request,
            provider=provider,
        )

    elif provider['protocol'] == 'websocket':
        from .rpc_backends import rpc_websocket

        return rpc_websocket.send_websocket(
            request=request,
            provider=provider,
        )

    else:
        raise Exception('unknown provider protocol: ' + str(provider['protocol']))


async def async_send(
    request: spec.RpcRequest,
    provider: typing.Optional[spec.ProviderSpec] = None,
) -> spec.RpcResponse:
    full_provider = rpc_provider.get_provider(provider)

    if isinstance(request, dict):
        response = await async_send_raw(request=request, provider=full_provider)
        if 'result' not in response and 'error' in response:
            raise Exception('RPC ERROR: ' + response['error']['message'])
        return response['result']

    elif isinstance(request, list):

        # chunk request
        request_chunks = chunk_request(request=request, provider=full_provider)

        logger = logging.getLogger()
        logger.info(
            'chunking batch of '
            + str(len(request))
            + ' calls into '
            + str(len(request_chunks))
            + ' chunks'
        )

        # send request chunks
        coroutines = []
        for request_chunk in request_chunks:
            coroutine = async_send_raw(
                request=request_chunk,
                provider=full_provider,
            )
            coroutines.append(coroutine)
        response_chunks = await asyncio.gather(*coroutines)

        # reorder chunks
        plural_response = reorder_response_chunks(response_chunks, request)

        return [subresponse['result'] for subresponse in plural_response]

    else:

        raise Exception('unknown request type: ' + str(type(request)))


@typing.overload
async def async_send_raw(
    request: spec.RpcSingularRequest, provider
) -> spec.RpcSingularResponseRaw:
    ...


@typing.overload
async def async_send_raw(
    request: spec.RpcPluralRequest, provider
) -> spec.RpcPluralResponseRaw:
    ...


async def async_send_raw(
    request: spec.RpcRequest,
    provider: spec.Provider,
) -> spec.RpcResponseRaw:

    _log_request(request=request, provider=provider)

    if provider['protocol'] == 'http':
        from .rpc_backends import rpc_http_async

        return await rpc_http_async.async_send_http(
            request=request,
            provider=provider,
        )

    elif provider['protocol'] == 'websocket':
        from .rpc_backends import rpc_websocket_async

        return await rpc_websocket_async.async_send_websocket(
            request=request,
            provider=provider,
        )

    else:
        raise Exception('unknown provider protocol: ' + str(provider['protocol']))


def _log_request(request: spec.RpcRequest, provider: spec.Provider) -> None:
    logger = logging.getLogger()
    if isinstance(request, dict):
        logger.info('singular request: ' + request['method'])
    elif isinstance(request, list):
        methods = [subrequest['method'] for subrequest in request]
        logger.info(
            'plural request: '
            + str(len(request))
            + ' calls, '
            + str(set(methods))
        )
    else:
        raise Exception('unknown request type: ' + str(type(request)))


def reorder_response_chunks(
    response_chunks: list[spec.RpcPluralResponseRaw],
    request: spec.RpcPluralRequest,
) -> spec.RpcPluralResponse:

    responses_by_id = {
        response['id']: response
        for response_chunk in response_chunks
        for response in response_chunk
    }
    return [responses_by_id[subrequest['id']] for subrequest in request]


#
# # chunking
#


def chunk_request(
    request: spec.RpcPluralRequest, provider: spec.Provider
) -> list[spec.RpcPluralRequest]:

    if provider['chunk_size'] is not None:
        return chunk_request_by_size(request, provider['chunk_size'])
    else:
        return [request]


def chunk_request_by_size(
    request: spec.RpcPluralRequest, chunk_size: int
) -> list[spec.RpcPluralRequest]:

    n_chunks = math.ceil(len(request) / chunk_size)
    return [
        request[slice(c * chunk_size, (c + 1) * chunk_size)]
        for c in range(n_chunks)
    ]


def chunk_request_by_method(request):
    raise NotImplementedError()


def chunk_request_by_block_range(request):
    raise NotImplementedError()




from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_eth_block_number(provider=None, decode_response=True):
    request = rpc_constructors.construct_eth_block_number()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_block_number(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_get_block_by_hash(
    block_hash,
    include_full_transactions=False,
    provider=None,
    decode_response=True,
    snake_case_response=True,
):
    request = rpc_constructors.construct_eth_get_block_by_hash(
        block_hash=block_hash,
        include_full_transactions=include_full_transactions,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_block_by_hash(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_eth_get_block_by_number(
    block_number,
    include_full_transactions=True,
    decode_response=True,
    provider=None,
    snake_case_response=True,
):
    request = rpc_constructors.construct_eth_get_block_by_number(
        block_number=block_number,
        include_full_transactions=include_full_transactions,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_block_by_number(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_eth_get_uncle_count_by_block_hash(
    block_hash, provider=None, decode_response=True
):
    request = rpc_constructors.construct_eth_get_uncle_count_by_block_hash(
        block_hash=block_hash,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_uncle_count_by_block_hash(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_get_uncle_count_by_block_number(
    block_number, provider=None, decode_response=True
):
    request = rpc_constructors.construct_eth_get_uncle_count_by_block_number(
        block_number=block_number,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_uncle_count_by_block_number(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_get_uncle_by_block_hash_and_index(
    block_hash,
    uncle_index,
    provider=None,
    decode_response=True,
    snake_case_response=True,
):
    request = rpc_constructors.construct_eth_get_uncle_by_block_hash_and_index(
        block_hash=block_hash,
        uncle_index=uncle_index,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_uncle_by_block_hash_and_index(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_eth_get_uncle_by_block_number_and_index(
    block_number,
    uncle_index,
    provider=None,
    decode_response=True,
    snake_case_response=True,
):
    request = (
        rpc_constructors.construct_eth_get_uncle_by_block_number_and_index(
            block_number=block_number,
            uncle_index=uncle_index,
        )
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_uncle_by_block_number_and_index(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )




from ctc import spec
from .. import rpc_request


def construct_eth_gas_price() -> spec.RpcRequest:
    return rpc_request.create('eth_gasPrice', [])


def construct_eth_accounts() -> spec.RpcRequest:
    return rpc_request.create('eth_accounts', [])


def construct_eth_sign(address, message) -> spec.RpcRequest:
    return rpc_request.create('eth_sign', [address, message])


def construct_eth_sign_transaction(
    from_address,
    data,
    to_address=None,
    gas=None,
    gas_price=None,
    value=None,
    nonce=None,
) -> spec.RpcRequest:

    parameters = {
        'from': from_address,
        'to': to_address,
        'gas': gas,
        'gasPrice': gas_price,
        'value': value,
        'data': data,
        'nonce': nonce,
    }
    parameters = {k: v for k, v in parameters.items() if v is not None}

    return rpc_request.create('eth_signTransaction', [parameters])


def construct_eth_send_transaction(
    from_address: spec.BinaryData,
    data: spec.BinaryData,
    to_address: spec.BinaryData = None,
    gas: spec.BinaryData = None,
    gas_price: spec.BinaryData = None,
    value: spec.BinaryData = None,
    nonce: spec.BinaryData = None,
) -> spec.RpcRequest:
    parameters = {
        'from': from_address,
        'to': to_address,
        'gas': gas,
        'gasPrice': gas_price,
        'value': value,
        'data': data,
        'nonce': nonce,
    }
    parameters = {k: v for k, v in parameters.items() if v is not None}

    return rpc_request.create('eth_sendTransaction', [parameters])


def construct_eth_send_raw_transaction(
    data: spec.BinaryData,
) -> spec.RpcRequest:
    return rpc_request.create('eth_sendRawTransaction', [data])




from .coracle_balances import *
from .coracle_deposits import *
from .coracle_fei import *
from .coracle_oracles import *
from .coracle_spec import *
from .coracle_stats import *
from .coracle_tokens import *



import copy

from . import abi_io
from .. import event_utils


#
# # summaries
#


async def async_summarize_contract_abi(
    contract_abi=None, contract_address=None
):
    if contract_abi is None:
        contract_abi = abi_io.async_get_contract_abi(
            contract_address=contract_address,
        )
    df = contract_abi_to_dataframe(contract_abi, human_readable=True)
    import IPython  # type: ignore

    IPython.display.display(df)


def summarize_contract_events(*, contract_abi=None, events=None):

    if events is None:
        events = get_contract_events(contract_abi=contract_abi)

    print(len(events), 'events:')
    for event in events.values():
        print('-', event['name'])
        for var in event['inputs']:
            if var['indexed']:
                index = 'topic'
            else:
                index = 'data'
            print('    - ' + var['name'] + ':', var['type'] + ',', index)


def get_contract_events(contract_abi=None, **abi_query):
    """get contract events by hash, as {event_hash: event_abi}"""

    if contract_abi is None:
        contract_abi = abi_io.get_contract_abi(**abi_query)

    return {
        binary.get_event_hash(event_abi=abi_item): abi_item
        for abi_item in contract_abi
        if abi_item['type'] == 'event'
    }


#
# # dataframes
#


def contract_abi_to_dataframe(contract_abi, human_readable):
    contract_abi = copy.deepcopy(contract_abi)
    for entry in contract_abi:

        if human_readable:

            if 'name' not in entry:
                entry['name'] = ''

            if 'inputs' in entry:
                inputs = []
                for input in entry['inputs']:
                    input_str = ''
                    input_str += input.get('type')
                    input_name = input.get('name')
                    if input_name is not None and input_name != '':
                        input_str += ' ' + input_name
                    inputs.append(input_str)
                entry['inputs'] = ', '.join(inputs)
            else:
                entry['inputs'] = ''

            if 'outputs' in entry:
                outputs = []
                for output in entry['outputs']:
                    output_str = ''
                    output_str += output.get('type')
                    output_name = output.get('name')
                    if output_name is not None and output_name != '':
                        output_str += ' ' + output_name
                    outputs.append(output_str)
                entry['outputs'] = ', '.join(outputs)
            else:
                entry['outputs'] = ''

            if 'anonymous' not in entry:
                entry['anonymous'] = ''

            if 'stateMutability' not in entry:
                entry['stateMutability'] = ''

    import pandas as pd

    df = pd.DataFrame(contract_abi)
    df = df.reindex(
        columns=[
            'type',
            'name',
            'inputs',
            'outputs',
            'stateMutability',
            'anonymous',
        ]
    )

    return df


def get_contract_events_dataframe(
    contract_abi, contract_name=None, contract_address=None, protocol_name=None
):
    event_abis = event_utils.async_get_contract_events(
        contract_abi=contract_abi
    )
    event_rows = []
    for event_hash, event_abi in event_abis.items():

        data_types = binary.get_event_data_types(event_abi=event_abi)
        data_names = binary.get_event_data_names(event_abi=event_abi)
        data_signature = [
            data_type + ' ' + data_name
            for data_type, data_name in zip(data_types, data_names)
        ]

        event_row = {
            'protocol_name': protocol_name,
            'contract_address': contract_address,
            'contract_name': contract_name,
            'event_name': event_abi['name'],
            'event_hash': event_hash,
            'topic1_type': None,
            'topic1_name': None,
            'topic2_type': None,
            'topic2_name': None,
            'topic3_type': None,
            'topic3_name': None,
            'data_signature': data_signature,
        }

        n_topics = 0
        for var in event_abi['inputs']:
            if var['indexed']:
                n_topics += 1
                event_row['topic' + str(n_topics) + '_type'] = var['type']
                event_row['topic' + str(n_topics) + '_name'] = var['name']

        event_rows.append(event_row)

    import pandas as pd

    return pd.DataFrame(event_rows)




from .rpc_block_digestors import *
from .rpc_log_digestors import *
from .rpc_mining_digestors import *
from .rpc_node_digestors import *
from .rpc_state_digestors import *
from .rpc_submission_digestors import *
from .rpc_transaction_digestors import *
from .rpc_whisper_digestors import *



import pytest

from ctc import rpc


@pytest.mark.asyncio
async def test_eth_block_number():
    result = await rpc.async_eth_block_number()
    assert isinstance(result, int)


@pytest.mark.asyncio
async def test_eth_get_block_by_hash():
    block_hash = (
        '0x767c2bfb3bdee3f78676c1285cd757bcd5d8c272cef2eb30d9733800a78c0b6d'
    )
    block = await rpc.async_eth_get_block_by_hash(block_hash)
    assert block['number'] == 12345


@pytest.mark.asyncio
async def test_eth_get_block_by_number():
    block = await rpc.async_eth_get_block_by_number(12345)
    assert (
        block['hash']
        == '0x767c2bfb3bdee3f78676c1285cd757bcd5d8c272cef2eb30d9733800a78c0b6d'
    )


@pytest.mark.asyncio
async def test_eth_get_uncle_count_by_block_hash():
    assert 2 == await rpc.async_eth_get_uncle_count_by_block_hash(
        '0x461336bad4949d91e253631a8f5df8412d258b59da9715ff638922f3c67353f8'
    )


@pytest.mark.asyncio
async def test_eth_get_uncle_count_by_block_number():
    assert 2 == await rpc.async_eth_get_uncle_count_by_block_number(222)


@pytest.mark.asyncio
async def test_eth_get_uncle_by_block_hash_and_index():
    block = await rpc.async_eth_get_uncle_by_block_hash_and_index(
        '0x461336bad4949d91e253631a8f5df8412d258b59da9715ff638922f3c67353f8', 1
    )
    assert (
        block['hash']
        == '0x4cf24f3784d19179965ad4f8396fce0f2bf2466d1e25c197021c0969b686f236'
    )


@pytest.mark.asyncio
async def test_eth_get_uncle_by_block_number_and_index():
    block = await rpc.async_eth_get_uncle_by_block_number_and_index(222, 1)
    assert (
        block['hash']
        == '0x4cf24f3784d19179965ad4f8396fce0f2bf2466d1e25c197021c0969b686f236'
    )




from __future__ import annotations

import typing
from typing_extensions import TypedDict

from ctc import spec


trade_fee = 0.003


class PoolTokensMetadata(TypedDict):
    x_address: spec.Address
    y_address: spec.Address
    x_symbol: str
    y_symbol: str
    x_decimals: int
    y_decimals: int


class PoolState(TypedDict):
    x_reserves: typing.Union[int, float]
    y_reserves: typing.Union[int, float]
    lp_total_supply: typing.Union[int, float]


class PoolStateByBlock(TypedDict):
    x_reserves: list[typing.Union[int, float]]
    y_reserves: list[typing.Union[int, float]]
    lp_total_supply: list[typing.Union[int, float]]




from ctc.protocols import uniswap_v2_utils

from ctc.cli import cli_utils
from ctc import rpc


def get_command_spec():
    return {
        'f': async_burns_command,
        'args': [
            {'name': 'pool'},
            {'name': '--blocks', 'kwargs': {'nargs': '+'}},
            {'name': '--output'},
            {'name': '--overwrite', 'kwargs': {'action': 'store_true'}},
        ],
    }


async def async_burns_command(pool, blocks, output, overwrite):

    if blocks is not None:
        start_block, end_block = await cli_utils.async_resolve_block_range(blocks)
    else:
        start_block = None
        end_block = None

    burns = await uniswap_v2_utils.async_get_pool_burns(
        pool,
        start_block=start_block,
        end_block=end_block,
    )
    cli_utils.output_data(burns, output, overwrite)

    await rpc.async_close_http_session()




# https://docs.chain.link/docs/faq/

import typing
from typing_extensions import TypedDict

from ctc import spec


_FeedReference = typing.Union[spec.Address, str]


class FeedRoundData(TypedDict):
    answer: int
    round_id: int
    timestamp: int


class FeedRoundDataNormalized(TypedDict):
    answer: float
    round_id: int
    timestamp: int


path_templates = {
    'feed': '{chainlink_view}/feeds/{feed}/{feed}__{start_block}_to_{end_block}.csv',
}





from ctc.protocols import uniswap_v2_utils


async def async_get_pool_swaps(
    pool_address,
    start_block=None,
    end_block=None,
    replace_symbols=False,
    normalize=True,
):
    return await uniswap_v2_utils.async_get_pool_swaps(
        pool_address=pool_address,
        start_block=start_block,
        end_block=end_block,
        replace_symbols=replace_symbols,
        normalize=normalize,
    )




import typing

import numpy as np
import pandas as pd

from ctc import spec


def keep_level(index: pd.MultiIndex, level: str) -> pd.Index:
    """drop all levels in a multiindex except for one"""
    for other_level in list(index.names):
        if other_level != level:
            index = index.droplevel(other_level)
    return index


def interpolate_series(
    series: spec.Series,
    start_index: typing.Optional[int] = None,
    end_index: typing.Optional[int] = None,
    pre_fill_value: typing.Any = None,
    level: typing.Optional[str] = None,
) -> spec.Series:

    # drop extra levels
    old_index = series.index
    if isinstance(old_index, pd.MultiIndex):
        if level is None:
            raise Exception('must specify which index level to use')
        series = series.copy()
        series.index = keep_level(old_index, level)

    # remove duplicate index values, keeping last value of each duplicate
    series = series[~series.index.duplicated(keep='last')]

    # build new index
    if start_index is None:
        start_index = series.index.values[0]
    if end_index is None:
        end_index = series.index.values[-1]
    new_index = np.arange(start_index, end_index + 1, 1, dtype=int)

    # create new series
    new_series = typing.cast(
        pd.Series, series.reindex(new_index, fill_value=pd.NA)
    )

    # insert pre fill value
    if start_index < series.index.values[0]:
        if pre_fill_value is None:
            raise Exception('for early start must specify pre_fill_value')
        new_series.iloc[0] = pre_fill_value
    elif start_index > series.index.values[-1]:
        # case: indicies start after the series ends
        new_series.iloc[0] = series.values[-1]
    elif start_index > series.index.values[0]:
        # fill in any initial values that were cut off
        fill_index = np.nonzero(series.index > start_index)[0][0] - 1  # type: ignore
        new_series.iloc[0] = series.iloc[fill_index]

    # interpolate values
    new_series = new_series.fillna(method='ffill')

    return new_series


def interpolate_dataframe(
    df: spec.DataFrame,
    start_index: typing.Optional[int] = None,
    end_index: typing.Optional[int] = None,
    level: typing.Optional[str] = None,
) -> spec.DataFrame:

    # drop extra levels
    old_index = df.index
    if isinstance(old_index, pd.MultiIndex):
        if level is None:
            raise Exception('must specify which index level to use')
        df = df.copy()
        df.index = keep_level(old_index, level)

    # remove duplicate index values, keeping last value of each duplicate
    df = df[~df.index.duplicated(keep='last')]

    # build new index
    if start_index is None:
        start_index = df.index.values[0]
    if end_index is None:
        end_index = df.index.values[-1]
    new_index = np.arange(start_index, end_index + 1, 1, dtype=int)

    # create new series
    new_df = df.reindex(new_index, fill_value=pd.NA)

    # fill in cut off values
    if start_index > df.index.values[0]:
        fill_index = np.nonzero(df.index > start_index)[0][0] - 1
        new_df.iloc[0] = df.iloc[fill_index]

    # interpolate values
    new_df = new_df.fillna(method='ffill')

    return new_df




from ctc import directory
from ctc import rpc


async def async_get_pool_id(pool_address, block=None):
    return await rpc.async_eth_call(
        to_address=pool_address,
        function_name='getPoolId',
        block_number=block,
    )


async def async_get_pool_address(pool_id, block=None, vault=None):
    if vault is None:
        vault = directory.get_address(name='Vault', label='balancer')

    pool = await rpc.async_eth_call(
        to_address=vault,
        function_name='getPool',
        function_parameters=[pool_id],
        block_number=block,
    )
    return pool[0]


async def async_get_pool_tokens(
    *, pool_address=None, pool_id=None, block=None, vault=None
):

    if vault is None:
        vault = directory.get_address(name='Vault', label='balancer')
    if pool_id is None:
        pool_id = await async_get_pool_id(pool_address)

    pool_tokens = await rpc.async_eth_call(
        to_address=vault,
        function_name='getPoolTokens',
        function_parameters=[pool_id],
        block_number=block,
        package_named_outputs=True,
    )
    return list(pool_tokens['tokens'])




import typing

from ctc import spec


T = typing.TypeVar('T', spec.DataFrame, spec.Series)


def bin_by_blocks(
    data: T,
    blocks: typing.Sequence[typing.SupportsInt],
    block_index_name: typing.Optional[str] = 'block_number',
) -> T:
    """TODO: deprecate
    - either:
        - refactor using pd.cut() https://stackoverflow.com/a/33761120
        - kill this whole function
    """
    import numpy as np
    import pandas as pd

    if block_index_name is not None and len(data.index.names) > 1:
        for index_name in data.index.names:
            if index_name != 'block_number':
                data.index = data.index.droplevel(index_name)

    standard_blocks = [int(block) for block in blocks]
    new_data = data.groupby(
        np.digitize(data.index.values, standard_blocks)
    ).sum()
    new_data = pd.Series(
        [0] * len(blocks), index=range(1, len(blocks) + 1)
    ).add(new_data, fill_value=0)
    # if new_data.index[0] < standard_blocks[0]:
    #     new_data = new_data.iloc[1:]
    new_data.index = pd.Index(standard_blocks)
    new_data.index.name = 'gte_this_block'  # type: ignore

    return new_data




from ctc import rpc

from .. import rari_abis


async def async_get_ctoken_comptroller(ctoken, block='latest'):
    return await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['comptroller'],
    )


async def async_get_ctoken_underlying(ctoken, block='latest'):
    return await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['underlying'],
    )


async def async_get_ctoken_irm(ctoken, block='latest'):
    return await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['interestRateModel'],
    )

